import inspect

from ..functions import CF

_VN = CF()


class DocstringRewriteMeta(type):
    """Modify docstrings at time of import.

    **Methodology**

    To do this, we intercede before the class is created and modify
    the docstrings of the attributes defined on the class.

    Inherited methods are also modified. We cannot simply modify the
    docstrings of inherited methods, because then the parent classes'
    methods will have the wrong docstring. Instead, we must actually
    copy the functions, and then modify the docstring.

    Special treatment is given to methods decorated with
    ``@property``, ``@staticmethod`` and ``@classmethod``, as well as
    user-defined decorations.

    .. versionadded:: (cfdm) 1.8.7.0

    """

    # Based on
    # http://www.jesshamrick.com/2013/04/17/rewriting-python-docstrings-with-a-metaclass/

    def __new__(cls, class_name, parents, attrs):
        """Combines docstring substitutions across the inheritance tree.

        That is, combines docstring substitutions from all classes in the
        inheritance tree.

        The value for a key that occurs in multiple classes will be taken
        from the class closest to the child class.

        """
        class_name_lower = class_name.lower()

        docstring_rewrite = {}

        for parent in parents[::-1]:
            parent_docstring_rewrite = getattr(
                parent, "_docstring_substitutions", None
            )
            if parent_docstring_rewrite is not None:
                docstring_rewrite.update(parent_docstring_rewrite(parent))
            else:
                parent_docstring_rewrite = getattr(
                    parent, "__docstring_substitutions__", None
                )
                if parent_docstring_rewrite is not None:
                    docstring_rewrite.update(parent_docstring_rewrite(None))

        class_docstring_rewrite = attrs.get(
            "__docstring_substitutions__", None
        )
        if class_docstring_rewrite is not None:
            docstring_rewrite.update(class_docstring_rewrite(None))

        special = DocstringRewriteMeta._docstring_special_substitutions()
        for key in special:
            if key in docstring_rewrite:
                raise ValueError(
                    f"Can't use {key!r} as a user-defined "
                    "docstring substitution."
                )

        # ------------------------------------------------------------
        # Find the package depth
        # ------------------------------------------------------------
        package_depth = 0
        for parent in parents[::-1]:
            parent_depth = getattr(parent, "_docstring_package_depth", None)
            if parent_depth is not None:
                package_depth = parent_depth(parent)
            else:
                parent_depth = getattr(
                    parent, "__docstring_package_depth__", None
                )
                if parent_depth is not None:
                    package_depth = parent_depth(None)

        class_depth = attrs.get("__docstring_package_depth__", None)
        if class_depth is not None:
            package_depth = class_depth(None)

        package_depth += 1

        module = attrs["__module__"].split(".")
        package_name = ".".join(module[:package_depth])

        # ------------------------------------------------------------
        # Find which methods to exclude
        # ------------------------------------------------------------
        method_exclusions = []
        for parent in parents[::-1]:
            parent_exclusions = getattr(
                parent, "_docstring_method_exclusions", None
            )
            if parent_exclusions is not None:
                method_exclusions.extend(parent_exclusions(parent))
            else:
                parent_exclusions = getattr(
                    parent, "__docstring_method_exclusions__", None
                )
                if parent_exclusions is not None:
                    method_exclusions.extend(parent_exclusions(None))

        class_exclusions = attrs.get("__docstring_method_exclusions__", None)
        if class_exclusions is not None:
            method_exclusions.extend(class_exclusions(None))

        method_exclusions = set(method_exclusions)

        for attr_name, attr in attrs.items():
            # Skip special methods that aren't functions
            if attr_name.startswith("__") and not inspect.isfunction(attr):
                continue

            # Skip methods without docstrings
            if not hasattr(attr, "__doc__"):
                continue

            if attr_name in method_exclusions:
                continue

            # @property
            if hasattr(attr, "fget"):
                # Note that here inspect.isroutine(attr) is False for
                # @property methods (but this is not the case for
                # properties in the parent classes).
                DocstringRewriteMeta._docstring_update(
                    package_name,
                    class_name,
                    class_name_lower,
                    attr,
                    attr_name,
                    docstring_rewrite,
                )
                continue

            # Still here?
            if not inspect.isroutine(attr):
                continue

            # Still here?
            is_classmethod = False
            is_staticmethod = False

            # Find out if the method is a classmethod (the
            # inspect.ismethod technique doesn't work for this class)
            if isinstance(attr, classmethod):
                is_classmethod = True
            elif isinstance(attr, staticmethod):
                is_staticmethod = True

            is_wrapped = hasattr(attr, "__wrapped__")
            if is_wrapped:
                wrapper = attr
                attr = attr.__wrapped__

            if is_classmethod or is_staticmethod:
                f = getattr(attr, "__func__", attr)

                # Copy the method
                attr = type(f)(
                    f.__code__,
                    f.__globals__,
                    f.__name__,
                    f.__defaults__,
                    f.__closure__,
                )

                # Make sure that the keyword argument defaults are set
                # correctly. In general they will be, but not if there
                # is a variable number of positional arguments, such
                # as in: def foo(self, *x, y=None)
                attr.__kwdefaults__ = f.__kwdefaults__

            # Update docstring
            DocstringRewriteMeta._docstring_update(
                package_name,
                class_name,
                class_name_lower,
                attr,
                attr_name,
                docstring_rewrite,
            )

            # Redecorate
            if is_classmethod:
                attrs[attr_name] = classmethod(attr)

            elif is_staticmethod:
                attrs[attr_name] = staticmethod(attr)

            elif is_wrapped:
                wrapper.__doc__ = attr.__doc__
                wrapper.__wrapped__ = attr
                attrs[attr_name] = wrapper

        # ------------------------------------------------------------
        # Now loop round the parent classes, copying any methods that
        # they override and rewriting those docstrings.
        # ------------------------------------------------------------
        for parent in parents:
            for attr_name in dir(parent):
                if attr_name in attrs:
                    # We already have this method from higher up in
                    # the method resolution order, so do not overwrite
                    # it and move on to to the next method.
                    continue

                # ----------------------------------------------------
                # Get the original method, copy it, update the
                # docstring, and put the mosfied copy back into the
                # parent class.
                # ----------------------------------------------------
                original_f = getattr(parent, attr_name)

                # Skip special methods that aren't functions
                if attr_name.startswith("__") and not inspect.isfunction(
                    original_f
                ):
                    continue

                if attr_name in method_exclusions:
                    continue

                is_classmethod = False
                is_staticmethod = False
                is_wrapped = False

                try:
                    if hasattr(original_f, "fget"):
                        # The original function is decorated with
                        # @property
                        attr = type(original_f)(
                            original_f.fget, original_f.fset, original_f.fdel
                        )
                    else:
                        if not inspect.isroutine(original_f):
                            continue

                        if inspect.ismethod(original_f):
                            is_classmethod = True
                        elif isinstance(
                            parent.__dict__.get(attr_name), staticmethod
                        ):
                            is_staticmethod = True

                        is_wrapped = hasattr(original_f, "__wrapped__")

                        f = getattr(original_f, "__func__", original_f)

                        # Copy the method
                        attr = type(f)(
                            f.__code__,
                            f.__globals__,
                            f.__name__,
                            f.__defaults__,
                            f.__closure__,
                        )
                        # Make sure that the keyword argument defaults
                        # are set correctly. In general they will be,
                        # but not if there is a variable number of
                        # positional arguments, such as in: def
                        # foo(self, *x, y=None)
                        attr.__kwdefaults__ = f.__kwdefaults__

                        if is_wrapped:
                            attr.__doc__ = original_f.__doc__

                    # Update the docstring
                    DocstringRewriteMeta._docstring_update(
                        package_name,
                        class_name,
                        class_name_lower,
                        attr,
                        attr_name,
                        docstring_rewrite,
                    )

                    # Register a classmethod
                    if is_classmethod:
                        attr = classmethod(attr)

                    # Register a classmethod
                    if is_staticmethod:
                        attr = staticmethod(attr)

                    if is_wrapped:
                        # Copy the wrapper and update its wrapped
                        # function
                        wrapper = type(original_f)(
                            original_f.__code__,
                            original_f.__globals__,
                            original_f.__name__,
                            original_f.__defaults__,
                            original_f.__closure__,
                        )

                        wrapper.__wrapped__ = attr
                        wrapper.__doc__ = attr.__doc__
                        attr = wrapper

                    # Put the modified method back into the parent
                    # class
                    attrs[attr_name] = attr

                except Exception:
                    pass
        #                    raise RuntimeError(str(error) + ': ' +
        #                                       '.'.join([parent.__name__,
        #                                                 attr_name]))

        # ------------------------------------------------------------
        # Rewrite the docstring of the class itself.
        #
        # The method is as follows:
        #
        # 1. If __doc__ contains substitutions then save the
        #    unsubstituted docstring in __doc_template__, rewrite the
        #    docstring, and save it in __doc__.
        #
        # 2. If __doc__ is not None and does not contain substitutions
        #    then set __doc_template___ to None.
        #
        # 3. If __doc__ is None then search back through the parent
        #    classes until you found one with a non-None __doc__ AND a
        #    non-None __doc_template__. If such a parent exists then
        #    copy its __doc_template__ to the child class's
        #    __doc_template__, rewrite it, and save the rewritten
        #    docstring to the child class's __doc__.
        #
        # ------------------------------------------------------------
        doc = attrs.get("__doc__")
        doc_template = None
        set_doc_template_to_None = False

        if doc is None:
            for parent in parents[::-1]:
                x = getattr(parent, "__doc__", None)
                if x is not None:
                    doc_template = getattr(parent, "__doc_template__", None)
                    if doc_template is not None:
                        break

            if doc_template is None:
                set_doc_template_to_None = True

        if doc_template is not None:
            doc = doc_template

        if doc is not None and "{{" in doc:
            doc_template = doc
            doc = DocstringRewriteMeta._docstring_update(
                package_name,
                class_name,
                class_name_lower,
                None,
                None,
                docstring_rewrite,
                class_docstring=doc,
            )
            attrs["__doc__"] = doc

            if set_doc_template_to_None:
                doc_template = None

        attrs["__doc_template__"] = doc_template

        # ------------------------------------------------------------
        # Create the class
        # ------------------------------------------------------------
        return super().__new__(cls, class_name, parents, attrs)

    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    @classmethod
    def _docstring_special_substitutions(cls):
        """Return the special docstring substitutions.

        ``{{class}}`` is replaced by the name of the class.

        ``{{class_lower}}`` is replaced by the name of the class
        convert to all lower case.

        ``{{package}}`` is replaced by the name of the package, as defined
        by the first N ``.`` (dot) separated fields of the class's
        `__module__` attribute, where is N determined by
        `_docstring_package_depth`.

        .. versionadded:: (cfdm) 1.8.7.0

        .. seealso:: `_docstring_package_depth`,
                     `_docstring_method_exclusions`,
                     `_docstring_substitutions`,
                     `__docstring_substitutions__`,
                     `__docstring_package_depth__`,
                     `__docstring_method_exclusions__`

        :Returns:

            `tuple`
                The special docstring substitution identifiers.

        """
        return ("{{class}}", "{{class_lower}}", "{{package}}", "{{VN}}")

    @staticmethod
    def _docstring_substitutions(cls):
        """Returns the substitutions that apply to methods of the class.

        Text to be replaced is specified as a key in the returned
        dictionary, with the replacement text defined by the corresponding
        value.

        Special docstring substitutions, as defined by a class's
        `_docstring_special_substitutions` method, may be used in the
        replacement text, and will be substituted as usual.

        Replacement text may contain other non-special substitutions.

        .. note:: The values are only checked once for embedded
                  non-special substitutions, so if the embedded
                  substitution itself contains a non-special substitution
                  then the latter will *not* be replaced. This restriction
                  is to prevent the possibility of infinite recursion.

        A key must be either a `str` or a `re.Pattern` object.

        If a key is a `str` then the corresponding value must be a string.

        If a key is a `re.Pattern` object then the corresponding value
        must be a string or a callable, as accepted by the
        `re.Pattern.sub` method.

        .. versionadded:: (cfdm) 1.8.7.0

        .. seealso:: `_docstring_special_substitutions`,
                     `_docstring_package_depth`,
                     `_docstring_method_exclusions`,
                     `__docstring_substitutions__`,
                     `__docstring_package_depth__`,
                     `__docstring_method_exclusions__`

        :Parameters:

            cls: class
                The class.

        :Returns:

            `dict`
                The docstring substitutions. A dictionary key matches text
                in the docstrings, with a corresponding value its
                replacement.

        """
        out = {}

        for klass in cls.__bases__[::-1]:
            d_s = getattr(klass, "_docstring_substitutions", None)
            if d_s is not None:
                out.update(d_s(klass))
            else:
                d_s = getattr(klass, "__docstring_substitutions__", None)
                if d_s is not None:
                    out.update(d_s(None))

        d_s = getattr(cls, "__docstring_substitutions__", None)
        if d_s is not None:
            out.update(d_s(None))

        return out

    @staticmethod
    def _docstring_package_depth(cls):
        """Returns the class {{package}} substitutions package depth.

        In docstrings, ``{{package}}`` is replaced by the name of the
        package, as defined by the first N+1 ``.`` (dot) separated fields
        of the class's `__module__` attribute.

        N defaults to 0, but may be set to any non-negative integer, M, by
        creating a `__docstring_package_depth__` method that returns M.

        .. versionadded:: (cfdm) 1.8.7.0

        .. seealso:: `_docstring_special_substitutions`,
                     `_docstring_substitutions`,
                     `_docstring_method_exclusions`,
                     `__docstring_substitutions__`,
                     `__docstring_package_depth__`,
                     `__docstring_method_exclusions__`

        :Parameters:

            cls: class
                The class.

        :Returns:

            `int`
                The package depth.

        """
        out = 0

        for klass in cls.__bases__[::-1]:
            d_s = getattr(klass, "_docstring_package_depth", None)
            if d_s is not None:
                out = d_s(klass)
            else:
                d_s = getattr(klass, "__docstring_package_depth__", None)
                if d_s is not None:
                    out = d_s(None)

        d_s = getattr(cls, "__docstring_package_depth__", None)
        if d_s is not None:
            out = d_s(None)

        return out

    @staticmethod
    def _docstring_method_exclusions(cls):
        """Returns method names excluded in the class substitutions.

        Exclusions for a class may be defined by creating a
        `__docstring_method_exclusions__` method that returns the sequence
        of names of methods to be excluded. These exclusions will also
        apply to any child classes.

        Exclusions may be defined for any reason, but in particular may
        be required if a method has a non-rewritable docstring. An
        example of method that has a non-rewritable docstring is when the
        method is a 'method_descriptor' object, such as `list.append`: any
        class that inherits such such a method will need to exclude it,
        unless it is explicitly overridden in the child class.

        .. versionadded:: (cfdm) 1.8.7.0

        .. seealso:: `_docstring_special_substitutions`,
                     `_docstring_substitutions`,
                     `_docstring_package_depth`,
                     `__docstring_substitutions__`,
                     `__docstring_package_depth__`,
                     `__docstring_method_exclusions__`

        :Parameters:

            cls: class
                The class.

        :Returns:

            `set`
                The names of the methods to exclude from the docstring
                substitution process.

        """
        out = ["_docstring_special_substitutions", "_docstring_package_depth"]

        for klass in cls.__bases__[::-1]:
            d_s = getattr(klass, "_docstring_method_exclusions", None)
            if d_s is not None:
                out.extend(d_s(klass))
            else:
                d_s = getattr(klass, "__docstring_method_exclusions__", None)
                if d_s is not None:
                    out.extend(d_s(None))

        d_s = getattr(cls, "__docstring_method_exclusions__", None)
        if d_s is not None:
            out.extend(d_s(None))

        return set(out)

    @classmethod
    def _docstring_update(
        cls,
        package_name,
        class_name,
        class_name_lower,
        f,
        method_name,
        config,
        class_docstring=None,
    ):
        """Performs docstring substitutions on a method at import time.

        .. versionadded:: (cfdm) 1.8.7.0

        :Parameters:

            package_name: `str`

            class_name: `str`

            f: class method

            method_name: `str`

            config: `dict`

        """
        if class_docstring is not None:
            doc = class_docstring
        else:
            doc = f.__doc__
            if doc is None or "{{" not in doc:
                return doc

        # ------------------------------------------------------------
        # Do general substitutions first
        # ------------------------------------------------------------
        for key, value in config.items():
            # Substitute the key for the value
            try:
                # Compiled regular expression substitution
                doc = key.sub(value, doc)
            except AttributeError:
                # String substitution
                doc = doc.replace(key, value)

        # ------------------------------------------------------------
        # Now do special substitutions
        # ------------------------------------------------------------
        # Insert the name of the package
        doc = doc.replace("{{package}}", package_name)

        # Insert the name of the class containing this method
        doc = doc.replace("{{class}}", class_name)

        # Insert the lower case name of the class containing this method
        doc = doc.replace("{{class_lower}}", class_name_lower)

        # Insert the CF version
        doc = doc.replace("{{VN}}", _VN)

        # ----------------------------------------------------------------
        # Set the rewritten docstring on the method
        # ----------------------------------------------------------------
        if class_docstring is None:
            f.__doc__ = doc

        return doc



from itertools import accumulate, product
from numbers import Number

import numpy as np

from ..netcdfindexer import netcdf_indexer
from .compressedarray import CompressedArray


class MeshArray(CompressedArray):
    """Abstract base class for data based on a UGRID connectivity array.

    .. versionadded:: (cfdm) 1.11.0.0

    """

    def __new__(cls, *args, **kwargs):
        """Store subarray classes.

        A child class must define its subarray classes in the
        `_Subarray` dictionary.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        instance = super().__new__(cls)
        instance._Subarray = {}
        return instance

    def __init__(
        self,
        connectivity=None,
        shape=None,
        compressed_dimensions=None,
        compression_type=None,
        start_index=None,
        cell_dimension=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            connectivity: array_like
                A 2-d integer array of indices that corresponds to a
                UGRID "edge_node_connectivity",
                "face_node_connectivity", or
                "face_face_connectivity" variable.

            shape: `tuple`
                The shape of the CF data model view of the
                connectivity array.

            {{init start_index: `int`}}

            {{init cell_dimension: `int`}}

            compression_type: `str`, optional
                The type of compression.

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            compressed_array=connectivity,
            shape=shape,
            compressed_dimensions=compressed_dimensions,
            compression_type=compression_type,
            source=source,
            copy=copy,
        )

        if source is not None:
            try:
                start_index = source.get_start_index(None)
            except AttributeError:
                start_index = None

            try:
                cell_dimension = source.get_cell_dimension(None)
            except AttributeError:
                cell_dimension = None

        if start_index is not None:
            self._set_component("start_index", start_index, copy=False)

        if cell_dimension is not None:
            self._set_component("cell_dimension", cell_dimension, copy=False)

    def __getitem__(self, indices):
        """Return a subspace of the uncompressed data.

        x.__getitem__(indices) <==> x[indices]

        Returns a subspace of the uncompressed array as an independent
        numpy array.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        from math import isnan

        # ------------------------------------------------------------
        # Method: Uncompress the entire array and then subspace it
        # ------------------------------------------------------------
        shape = self.shape
        # Find out whether the shape of the uncompressed array is
        # currently known, or not.
        known_shape = not any(map(isnan, shape))

        if known_shape:
            # The shape of the uncompressed array is already known, so
            # we can initialise the un-sliced uncompressed array.
            u = np.ma.empty(self.shape, dtype=self.dtype)

        Subarray = self.get_Subarray()
        subarray_kwargs = {
            **self.conformed_data(),
            **self.subarray_parameters(),
        }

        for u_indices, u_shape, c_indices, _ in zip(*self.subarrays()):
            subarray = Subarray(
                indices=c_indices,
                shape=u_shape,
                **subarray_kwargs,
            )
            if known_shape:
                # The shape of the uncompressed array is already
                # known, so we have an uninitialised output array in
                # which to copy this subarray.
                u[u_indices] = subarray[...]
            else:
                # The shape of the uncompressed array is not currently
                # known, so we don't have an uninitialised output
                # array in which to copy this subarray. In this case
                # there must be (and so we are assuming that there is)
                # only one pass through this loop, so we can simply
                # set the output array to be this subarray.
                u = subarray[...]

        if not known_shape:
            # The shape of the uncompressed array was not known, but
            # it is known now that the subarray has been uncompressed.
            # Therefore we can now store the uncompressed shape for
            # future reference.
            self._set_component("shape", u.shape, copy=False)

        u = netcdf_indexer(
            u,
            mask=False,
            unpack=False,
            always_masked_array=False,
            orthogonal_indexing=True,
            copy=False,
        )
        return u[indices]

    @property
    def dtype(self):
        """The data-type of the uncompressed data.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        return self._get_compressed_Array().dtype

    def get_cell_dimension(self, default=ValueError()):
        """Return the position of the cell dimension.

        In UGRID, this is provided by one of the the "edge_dimension",
        "face_dimension" or "volume_dimension" variables.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            default: optional
                Return the value of the *default* parameter if there
                is no cell dimension position.

                {{default Exception}}

        :Returns:

            `int`
                The position of the cell dimension.

        """
        return self._get_component("cell_dimension", default)

    def get_start_index(self, default=ValueError()):
        """Return the start index.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            default: optional
                Return the value of the *default* parameter if there
                is no start index.

                {{default Exception}}

        :Returns:

            `int`
                The start index.

        """
        return self._get_component("start_index", default)

    def subarray_parameters(self):
        """Non-data parameters required by the `Subarray` class.

        .. versionadded:: (cfdm) 1.11.0.0

        :Returns:

            `dict`
                The parameters, with keys ``'compressed_dimensions'``,
                ``'start_index'``, and ``'cell_dimension'``.

        """
        return {
            **super().subarray_parameters(),
            "start_index": self.get_start_index(),
            "cell_dimension": self.get_cell_dimension(),
        }

    def subarray_shapes(self, shapes):
        """Create the subarray shapes along each uncompressed dimension.

        Note that the output is independent of the *shapes* parameter,
        because each dimension of the compressed data corresponds to a
        unique dimension of the uncompressed data.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `subarray`

        :Parameters:

            {{subarray_shapes chunks: `int`, sequence, `dict`, or `str`, optional}}

        :Returns:

            `list`
                The subarray sizes along each uncompressed dimension.

        **Examples**

        >>> a.shape
        (4, 4)
        >>> a.compressed_dimensions()
        {0: (0,), 1: (1,)}
        >>> a.subarray_shapes(-1)
        [(4,), (4)]
        >>> a.subarray_shapes("auto")
        [(4,), (4)]
        >>> a.subarray_shapes("60B")
        [(4,), (4)]

        """
        if shapes == -1:
            return [(size,) for size in self.shape]

        u_dims = self.get_compressed_axes()

        if isinstance(shapes, (str, Number)):
            return [
                (size,) if i in u_dims else shapes
                for i, size in enumerate(self.shape)
            ]

        if isinstance(shapes, dict):
            shapes = [
                shapes[i] if i in shapes else None for i in range(self.ndim)
            ]
        elif len(shapes) != self.ndim:
            raise ValueError(
                f"Wrong number of 'shapes' elements in {shapes}: "
                f"Got {len(shapes)}, expected {self.ndim}"
            )

        # chunks is a sequence
        return [
            (size,) if i in u_dims else c
            for i, (size, c) in enumerate(zip(self.shape, shapes))
        ]

    def subarrays(self, shapes=-1):
        """Return descriptors for every subarray.

        These descriptors are used during subarray decompression.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            {{subarrays chunks: ``-1`` or sequence, optional}}

        :Returns:

             4-`tuple` of iterators
                Each iterable iterates over a particular descriptor
                from each subarray:

                1. The indices of the uncompressed array that
                   correspond to each subarray.

                2. The shape of each uncompressed subarray.

                3. The indices of the compressed array that correspond
                   to each subarray.

                4. The location of each subarray on the uncompressed
                   dimensions.

        **Examples**

        A compressed array with shape (4, 4) that has an original
        UGRID connectivity array with shape (4, 2).

        >>> u_indices, u_shapes, c_indices, locations = x.subarrays()
        >>> for i in u_indices:
        ...    print(i)
        ...
        (slice(0, 4, None), slice(0, 4, None))
        >>> for i in u_shapes
        ...    print(i)
        ...
        (4, 4)
        >>> for i in c_indices:
        ...    print(i)
        ...
        (slice(0, 4, None), slice(0, 2, None))
        >>> for i in locations:
        ...    print(i)
        ...
        (0, 0)

        """
        dims = self.compressed_dimensions().keys()

        shapes = self.subarray_shapes(shapes)

        # The indices of the uncompressed array that correspond to
        # each subarray, the shape of each uncompressed subarray, and
        # the location of each subarray
        locations = []
        u_shapes = []
        u_indices = []
        for d, (size, c) in enumerate(zip(self.shape, shapes)):
            if d in dims:
                locations.append((0,))
                u_shapes.append((size,))
                u_indices.append((slice(None),))
            else:
                # Note: c can not be (nan,) when d is not a compressed
                #       dimension
                locations.append([i for i in range(len(c))])
                u_shapes.append(c)

                c = tuple(accumulate((0,) + c))
                u_indices.append([slice(i, j) for i, j in zip(c[:-1], c[1:])])

        # The indices of the compressed array that correspond to each
        # subarray
        if 0 in dims:
            c_indices = [(slice(None),)]
        else:
            c_indices = [u_indices[0]]

        c_indices.append((slice(None),))

        return (
            product(*u_indices),
            product(*u_shapes),
            product(*c_indices),
            product(*locations),
        )



import atexit
import datetime
import faulthandler
import os
import tempfile
import unittest
from pathlib import PurePath

import netCDF4

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm
from cfdm.read_write.netcdf.netcdfwrite import AggregationError

n_tmpfiles = 5
tmpfiles = [
    tempfile.mkstemp("_test_CFA.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
(
    tmpfile1,
    tmpfile2,
    nc_file,
    cfa_file,
    cfa_file2,
) = tmpfiles


def _remove_tmpfiles():
    """Try to remove defined temporary files by deleting their paths."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)


class CFATest(unittest.TestCase):
    """Unit test for aggregation variables."""

    netcdf3_fmts = [
        "NETCDF3_CLASSIC",
        "NETCDF3_64BIT",
        "NETCDF3_64BIT_OFFSET",
        "NETCDF3_64BIT_DATA",
    ]
    netcdf4_fmts = ["NETCDF4", "NETCDF4_CLASSIC"]
    netcdf_fmts = netcdf3_fmts + netcdf4_fmts

    aggregation_value = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "aggregation_value.nc"
    )
    f0 = cfdm.example_field(0)

    def test_CFA_fmt(self):
        """Test the cfdm.read 'fmt' keyword with cfa."""
        f = self.f0
        cfdm.write(f, tmpfile1)
        f = cfdm.read(tmpfile1, cfa_write="field")[0]

        for fmt in self.netcdf_fmts:
            cfdm.write(f, cfa_file, fmt=fmt, cfa="field")
            g = cfdm.read(cfa_file)
            self.assertEqual(len(g), 1)
            self.assertTrue(f.equals(g[0]))

    def test_CFA_multiple_fragments(self):
        """Test aggregation variables with more than one fragment."""
        f = self.f0

        cfdm.write(f[:2], tmpfile1)
        cfdm.write(f[2:], tmpfile2)

        a = cfdm.read(tmpfile1, cfa_write="field")[0]
        b = cfdm.read(tmpfile2, cfa_write="field")[0]
        a = cfdm.Field.concatenate([a, b], axis=0)

        cfdm.write(a, nc_file)
        cfdm.write(a, cfa_file, cfa="field")

        n = cfdm.read(nc_file)
        c = cfdm.read(cfa_file)
        self.assertEqual(len(n), 1)
        self.assertEqual(len(c), 1)
        self.assertTrue(c[0].equals(f))
        self.assertTrue(n[0].equals(c[0]))

    def test_CFA_strict(self):
        """Test 'strict' option to the cfdm.write 'cfa' keyword."""
        f = self.f0

        # By default, can't write in-memory arrays as aggregation
        # variables
        with self.assertRaises(AggregationError):
            cfdm.write(f, cfa_file, cfa="field")

        # The previous line should have deleted the output file
        self.assertFalse(os.path.exists(cfa_file))

        cfdm.write(f, nc_file, cfa={"constructs": "field", "strict": False})
        g = cfdm.read(nc_file, cfa_write="field")
        self.assertEqual(len(g), 1)
        self.assertTrue(g[0].equals(f))

        cfdm.write(g, cfa_file, cfa={"constructs": "field", "strict": True})
        g = cfdm.read(cfa_file)
        self.assertEqual(len(g), 1)
        self.assertTrue(g[0].equals(f))

    def test_CFA_uri_0(self):
        """Test aggregation 'uri' option to cfdm.write."""
        f = self.f0
        cfdm.write(f, tmpfile1)
        f = cfdm.read(tmpfile1, cfa_write="field")[0]

        absuri_filename = PurePath(os.path.abspath(tmpfile1)).as_uri()
        reluri_filename = os.path.basename(tmpfile1)

        for uri, filename in zip(
            ("absolute", "relative"), (absuri_filename, reluri_filename)
        ):
            cfdm.write(
                f,
                cfa_file,
                cfa={"constructs": "field", "uri": uri},
            )

            nc = netCDF4.Dataset(cfa_file, "r")
            cfa_location = nc.variables["cfa_location"]
            self.assertEqual(cfa_location[...], filename)
            nc.close()

            g = cfdm.read(cfa_file)
            self.assertEqual(len(g), 1)
            g = g[0]
            self.assertTrue(f.equals(g))
            self.assertEqual(
                g.data.get_filenames(normalise=False), set([filename])
            )

    def test_CFA_uri_1(self):
        """Test aggregation 'uri=default' option to cfdm.write."""
        f = self.f0
        cfdm.write(f, tmpfile1)
        f = cfdm.read(tmpfile1, cfa_write="field")[0]

        absuri_filename = PurePath(os.path.abspath(tmpfile1)).as_uri()
        reluri_filename = os.path.basename(tmpfile1)

        for uri, filename in zip(
            ("absolute", "relative"), (absuri_filename, reluri_filename)
        ):
            cfdm.write(
                f,
                cfa_file,
                cfa={"constructs": "field", "uri": uri},
            )

            g = cfdm.read(cfa_file)[0]
            cfdm.write(
                g,
                cfa_file2,
                cfa="field",
            )

            nc = netCDF4.Dataset(cfa_file2, "r")
            cfa_location = nc.variables["cfa_location"]
            self.assertEqual(cfa_location[...], filename)
            nc.close()

    def test_CFA_constructs(self):
        """Test aggregation 'constructs' option to cfdm.write."""
        f = cfdm.example_field(1)
        f.del_construct("time")
        f.del_construct("long_name=Grid latitude name")
        cfdm.write(f, tmpfile1)
        f = cfdm.read(tmpfile1, cfa_write="all")[0]

        # No constructs
        cfdm.write(f, tmpfile2, cfa={"constructs": []})
        nc = netCDF4.Dataset(tmpfile2, "r")
        for var in nc.variables.values():
            attrs = var.ncattrs()
            self.assertNotIn("aggregated_dimensions", attrs)
            self.assertNotIn("aggregated_data", attrs)

        nc.close()

        # Field construct
        cfdm.write(f, tmpfile2, cfa={"constructs": "field"})
        nc = netCDF4.Dataset(tmpfile2, "r")
        for ncvar, var in nc.variables.items():
            attrs = var.ncattrs()
            if ncvar in ("ta",):
                self.assertFalse(var.ndim)
                self.assertIn("aggregated_dimensions", attrs)
                self.assertIn("aggregated_data", attrs)
            else:
                self.assertNotIn("aggregated_dimensions", attrs)
                self.assertNotIn("aggregated_data", attrs)

        nc.close()

        # Dimension construct
        for constructs in (
            "dimension_coordinate",
            ["dimension_coordinate"],
            {"dimension_coordinate": None},
            {"dimension_coordinate": 1},
        ):
            cfdm.write(f, tmpfile2, cfa={"constructs": constructs})
            nc = netCDF4.Dataset(tmpfile2, "r")
            for ncvar, var in nc.variables.items():
                attrs = var.ncattrs()
                if ncvar in (
                    "x",
                    "x_bnds",
                    "y",
                    "y_bnds",
                    "atmosphere_hybrid_height_coordinate",
                    "atmosphere_hybrid_height_coordinate_bounds",
                ):
                    self.assertFalse(var.ndim)
                    self.assertIn("aggregated_dimensions", attrs)
                    self.assertIn("aggregated_data", attrs)
                else:
                    self.assertNotIn("aggregated_dimensions", attrs)
                    self.assertNotIn("aggregated_data", attrs)

            nc.close()

        # Dimension and auxiliary constructs
        for constructs in (
            ["dimension_coordinate", "auxiliary_coordinate"],
            {"dimension_coordinate": None, "auxiliary_coordinate": 2},
        ):
            cfdm.write(f, tmpfile2, cfa={"constructs": constructs})
            nc = netCDF4.Dataset(tmpfile2, "r")
            for ncvar, var in nc.variables.items():
                attrs = var.ncattrs()
                if ncvar in (
                    "x",
                    "x_bnds",
                    "y",
                    "y_bnds",
                    "atmosphere_hybrid_height_coordinate",
                    "atmosphere_hybrid_height_coordinate_bounds",
                    "latitude_1",
                    "longitude_1",
                ):
                    self.assertFalse(var.ndim)
                    self.assertIn("aggregated_dimensions", attrs)
                    self.assertIn("aggregated_data", attrs)
                else:
                    self.assertNotIn("aggregated_dimensions", attrs)
                    self.assertNotIn("aggregated_data", attrs)

            nc.close()

    def test_CFA_scalar(self):
        """Test scalar aggregation variable."""
        f = self.f0
        f = f[0, 0].squeeze()
        cfdm.write(f, tmpfile1)
        g = cfdm.read(tmpfile1, cfa_write="field")[0]
        cfdm.write(g, cfa_file, cfa="field")
        h = cfdm.read(cfa_file)[0]
        self.assertTrue(h.equals(f))

    def test_CFA_value(self):
        """Test the value fragment array variable."""
        write = True
        for aggregation_value_file in (self.aggregation_value, cfa_file):
            f = cfdm.read(aggregation_value_file, cfa_write="all")
            self.assertEqual(len(f), 1)
            f = f[0]
            fa = f.field_ancillary()
            self.assertEqual(fa.shape, (12,))
            self.assertEqual(fa.data.chunks, ((3, 9),))
            self.assertEqual(
                fa.data.nc_get_aggregation_fragment_type(), "unique_value"
            )
            self.assertEqual(
                fa.data.nc_get_aggregated_data(),
                {
                    "map": "fragment_map_uid",
                    "unique_value": "fragment_value_uid",
                },
            )

            nc = netCDF4.Dataset(aggregation_value_file, "r")
            fragment_value_uid = nc.variables["fragment_value_uid"][...]
            nc.close()

            self.assertTrue((fa[:3].array == fragment_value_uid[0]).all())
            self.assertTrue((fa[3:].array == fragment_value_uid[1]).all())

            if write:
                cfdm.write(f, cfa_file)
                write = False

    def test_CFA_cfa(self):
        """Test the cfdm.write 'cfa' keyword."""
        f = self.f0
        cfdm.write(f, tmpfile1)
        f = cfdm.read(tmpfile1, cfa_write="field")[0]
        cfdm.write(f, tmpfile2, cfa="field")
        g = cfdm.read(tmpfile2, cfa_write="field")[0]

        # Default of cfa="auto" - check that aggregation variable
        # gets written
        cfdm.write(g, cfa_file)
        nc = netCDF4.Dataset(cfa_file, "r")
        self.assertIsNotNone(
            getattr(nc.variables["q"], "aggregated_data", None)
        )
        nc.close()

        cfdm.write(g, cfa_file, cfa={"constructs": {"auto": 2}})
        nc = netCDF4.Dataset(cfa_file, "r")

        self.assertIsNotNone(
            getattr(nc.variables["q"], "aggregated_data", None)
        )
        nc.close()

        cfdm.write(
            g,
            cfa_file,
            cfa={
                "constructs": ["auto", "dimension_coordinate"],
                "strict": False,
            },
        )
        nc = netCDF4.Dataset(cfa_file, "r")
        for ncvar in ("q", "lat", "lon"):
            self.assertIsNotNone(
                getattr(nc.variables[ncvar], "aggregated_data", None)
            )

        nc.close()

        # Check bad values of cfa
        for cfa in (False, True, (), []):
            with self.assertRaises(ValueError):
                cfdm.write(g, cfa_file, cfa=cfa)

    def test_CFA_subspace(self):
        """Test the writing subspaces of aggregations."""
        f = self.f0

        cfdm.write(f[:2], tmpfile1)
        cfdm.write(f[2:], tmpfile2)

        a = cfdm.read(tmpfile1, cfa_write="field")[0]
        b = cfdm.read(tmpfile2, cfa_write="field")[0]
        c = cfdm.Field.concatenate([a, b], axis=0)

        cfdm.write(c, cfa_file, cfa="field")

        f = cfdm.read(cfa_file, cfa_write="field")[0]
        cfdm.write(f[:2], cfa_file2, cfa="field")
        g = cfdm.read(cfa_file2)[0]
        self.assertTrue(g.equals(a))

        cfdm.write(f[2:], cfa_file2, cfa="field")
        g = cfdm.read(cfa_file2)[0]
        self.assertTrue(g.equals(b))

        # Can't straddle Dask chunks
        with self.assertRaises(AggregationError):
            cfdm.write(f[1:3], cfa_file2, cfa="field")


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)



import datetime
import faulthandler
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm

# Create test domain topology object
c = cfdm.DomainTopology()
c.set_properties({"long_name": "Maps every face to its corner nodes"})
c.nc_set_variable("Mesh2_face_nodes")
data = cfdm.Data(
    [[2, 3, 1, 0], [6, 7, 3, 2], [1, 3, 8, -99]],
    dtype="i4",
)
data.masked_values(-99, inplace=True)
c.set_data(data)
c.set_cell("face")


class DomainTopologyTest(unittest.TestCase):
    """Unit test for the DomainTopology class."""

    d = c

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_DomainTopology__repr__str__dump(self):
        """Test all means of DomainTopology inspection."""
        d = self.d
        self.assertEqual(repr(d), "<DomainTopology: cell:face(3, 4) >")
        self.assertEqual(str(d), "cell:face(3, 4) ")
        self.assertEqual(
            d.dump(display=False),
            """Domain Topology: cell:face
    long_name = 'Maps every face to its corner nodes'
    Data(3, 4) = [[2, ..., --]]""",
        )

    def test_DomainTopology_copy(self):
        """Test the copy of DomainTopology."""
        d = self.d
        self.assertTrue(d.equals(d.copy()))

    def test_DomainTopology_data(self):
        """Test the data of DomainTopology."""
        d = self.d
        self.assertEqual(d.ndim, 1)

    def test_DomainTopology_cell(self):
        """Test the 'cell' methods of DomainTopology."""
        d = self.d.copy()
        self.assertTrue(d.has_cell())
        self.assertEqual(d.get_cell(), "face")
        self.assertEqual(d.del_cell(), "face")
        self.assertFalse(d.has_cell())
        self.assertIsNone(d.get_cell(None))
        self.assertIsNone(d.del_cell(None))

        with self.assertRaises(ValueError):
            d.get_cell()

        with self.assertRaises(ValueError):
            d.set_cell("bad value")

        self.assertIsNone(d.set_cell("face"))
        self.assertTrue(d.has_cell())
        self.assertEqual(d.get_cell(), "face")

    def test_DomainTopology_transpose(self):
        """Test the 'transpose' method of DomainTopology."""
        d = self.d.copy()
        e = d.transpose()
        self.assertTrue(d.equals(e))
        self.assertIsNone(d.transpose(inplace=True))

        for axes in ([1], [1, 0], [3]):
            with self.assertRaises(ValueError):
                d.transpose(axes)

    def test_DomainTopology_normalise(self):
        """Test the 'normalise' method of DomainTopology."""
        # Face cells
        data = cfdm.Data(
            [[1, 4, 5, 2], [4, 10, 1, -99], [122, 123, 106, 105]],
            mask_value=-99,
        )
        d = cfdm.DomainTopology(cell="face", data=data)

        n = np.ma.array([[0, 2, 3, 1], [2, 4, 0, -99], [7, 8, 6, 5]])
        n[1, -1] = np.ma.masked
        d0 = d.normalise().array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())
        d1 = d.normalise(start_index=1).array
        self.assertEqual(d1.shape, n.shape)
        self.assertTrue((d1.mask == n.mask).all())
        self.assertTrue((d1 == n + 1).all())

        e = d.copy()
        self.assertIsNone(e.normalise(inplace=True))
        d0 = e.normalise().array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())

        d.data[:, -1] = np.ma.masked
        n = np.ma.array([[0, 1, 2, -99], [1, 3, 0, -99], [5, 6, 4, -99]])
        n[:, -1] = np.ma.masked

        d0 = d.normalise().array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())

        n = n[:, :-1]
        d0 = d.normalise(remove_empty_columns=True).array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())

        # Point cells
        data = cfdm.Data(
            [[4, 1, 10, 125], [1, 4, -99, -99], [125, 4, -99, -99]],
            mask_value=-99,
        )
        d = cfdm.DomainTopology(cell="point", data=data)

        n = np.ma.array([[0, 1, 2, -99], [1, 0, -99, -99], [2, 0, -99, -99]])
        n = np.ma.where(n == -99, np.ma.masked, n)

        d0 = d.normalise().array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())
        d1 = d.normalise(start_index=1).array
        self.assertEqual(d1.shape, n.shape)
        self.assertTrue((d1.mask == n.mask).all())
        self.assertTrue((d1 == n + 1).all())

        e = d.copy()
        self.assertIsNone(e.normalise(inplace=True))
        d0 = e.normalise().array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())
        d1 = e.normalise(start_index=1).array
        self.assertEqual(d1.shape, n.shape)
        self.assertTrue((d1.mask == n.mask).all())
        self.assertTrue((d1 == n + 1).all())

        self.assertIsNone(e.normalise(start_index=1, inplace=True))
        d0 = e.normalise().array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())
        d1 = e.normalise(start_index=1).array
        self.assertEqual(d1.shape, n.shape)
        self.assertTrue((d1.mask == n.mask).all())
        self.assertTrue((d1 == n + 1).all())

        d.data[:, -1] = np.ma.masked
        n = np.ma.array([[0, 1, -99, -99], [1, 0, -99, -99], [2, 0, -99, -99]])
        n = np.ma.where(n == -99, np.ma.masked, n)

        d0 = d.normalise().array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())

        d.data[:, 1:] = np.ma.masked
        n = np.ma.array(
            [[0, -99, -99, -99], [1, -99, -99, -99], [2, -99, -99, -99]]
        )
        n = np.ma.where(n == -99, np.ma.masked, n)
        d0 = d.normalise().array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())

        n = np.ma.array([[0], [1], [2]])
        d0 = d.normalise(remove_empty_columns=True).array
        self.assertEqual(d0.shape, n.shape)
        self.assertTrue((d0.mask == n.mask).all())
        self.assertTrue((d0 == n).all())


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)



from . import core, mixin


class NodeCountProperties(
    mixin.NetCDFVariable,
    mixin.Properties,
    mixin.Files,
    core.abstract.Properties,
):
    """Properties for a netCDF node count variable.

    **NetCDF interface**

    {{netCDF variable}}

    .. versionadded:: (cfdm) 1.8.0

    """

    def __init__(self, properties=None, source=None, copy=True):
        """**Initialisation**

        :Parameters:

            {{init properties: `dict`, optional}}

                *Parameter example:*
                  ``properties={'long_name': 'number of nodes for each
                  geometry'}``

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(properties=properties, source=source, copy=copy)

        self._initialise_netcdf(source)
        self._initialise_original_filenames(source)

    def dump(
        self,
        display=True,
        _key=None,
        _title=None,
        _create_title=True,
        _prefix="",
        _level=0,
        _omit_properties=None,
    ):
        """A full description of the node count variable.

        Returns a description of all properties.

        .. versionadded:: (cfdm) 1.8.0

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

        :Returns:

            {{returns dump}}

        """
        if _create_title and _title is None:
            _title = "Node Count: " + self.identity(default="")

        return super().dump(
            display=display,
            _key=_key,
            _omit_properties=_omit_properties,
            _prefix=_prefix,
            _level=_level,
            _title=_title,
            _create_title=_create_title,
        )



import numpy as np

from . import Constructs, Domain, abstract, mixin


class Field(mixin.FieldDomain, abstract.PropertiesData):
    """A field construct of the CF data model.

    The field construct is central to the CF data model, and includes
    all the other constructs. A field corresponds to a CF-netCDF data
    variable with all of its metadata. All CF-netCDF elements are
    mapped to a field construct or some element of the CF field
    construct. The field construct contains all the data and metadata
    which can be extracted from the file using the CF conventions.

    The field construct consists of a data array and the definition of
    its domain (that describes the locations of each cell of the data
    array), field ancillary constructs containing metadata defined
    over the same domain, and cell method constructs to describe how
    the cell values represent the variation of the physical quantity
    within the cells of the domain. The domain is defined collectively
    by the following constructs of the CF data model: domain axis,
    dimension coordinate, auxiliary coordinate, cell measure,
    coordinate reference and domain ancillary constructs. All of the
    constructs contained by the field construct are optional.

    The field construct also has optional properties to describe
    aspects of the data that are independent of the domain. These
    correspond to some netCDF attributes of variables (e.g. units,
    long_name and standard_name), and some netCDF global file
    attributes (e.g. history and institution).

    .. versionadded:: (cfdm) 1.7.0

    """

    # ----------------------------------------------------------------
    # Define the base of the identity keys for each construct type
    # ----------------------------------------------------------------
    _construct_key_base = {
        "auxiliary_coordinate": "auxiliarycoordinate",
        "cell_connectivity": "cellconnectivity",
        "cell_measure": "cellmeasure",
        "cell_method": "cellmethod",
        "coordinate_reference": "coordinatereference",
        "dimension_coordinate": "dimensioncoordinate",
        "domain_ancillary": "domainancillary",
        "domain_axis": "domainaxis",
        "domain_topology": "domaintopology",
        "field_ancillary": "fieldancillary",
    }

    def __new__(cls, *args, **kwargs):
        """Store component classes.

        .. versionadded:: (cfdm) 1.7.0

        """
        instance = super().__new__(cls)
        instance._Constructs = Constructs
        instance._Domain = Domain
        return instance

    def __init__(
        self, properties=None, source=None, copy=True, _use_data=True
    ):
        """**Initialisation**

        :Parameters:

            {{init properties: `dict`, optional}}

                *Parameter example:*
                   ``properties={'standard_name': 'air_temperature'}``

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        # To avoid mutable default argument (an anti-pattern) of properties={}
        if properties is None:
            properties = {}

        super().__init__(
            properties=properties, source=source, copy=copy, _use_data=False
        )

        if source is not None:
            # Initialise constructs and the data from the source
            # parameter
            try:
                constructs = source.constructs
            except AttributeError:
                constructs = None

            try:
                data = source.get_data(default=None)
            except AttributeError:
                data = None

            try:
                data_axes = source.get_data_axes(default=None)
            except AttributeError:
                data_axes = None

            if constructs is not None and (copy or not _use_data):
                constructs = constructs.copy(data=_use_data)
        else:
            constructs = None
            data = None
            data_axes = None

        if constructs is None:
            constructs = self._Constructs(**self._construct_key_base)

        self._set_component("constructs", constructs, copy=False)

        if data is not None and _use_data:
            self.set_data(data, data_axes, copy=copy)
        elif data_axes is not None:
            self.set_data_axes(data_axes)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def construct_type(self):
        """Return a description of the construct type.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `str`
                The construct type.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.construct_type
        'field'

        """
        return "field"

    @property
    def constructs(self):
        """Return the metadata constructs.

        .. versionadded:: (cfdm) 1.7.0

        :Returns:

            `Constructs`
                The constructs.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> print(f.constructs)
        Constructs:
        {'cellmethod0': <{{repr}}CellMethod: area: mean>,
         'dimensioncoordinate0': <{{repr}}DimensionCoordinate: latitude(5) degrees_north>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: longitude(8) degrees_east>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: time(1) days since 2018-12-01 >,
         'domainaxis0': <{{repr}}DomainAxis: size(5)>,
         'domainaxis1': <{{repr}}DomainAxis: size(8)>,
         'domainaxis2': <{{repr}}DomainAxis: size(1)>}

        """
        return self._get_component("constructs")

    @property
    def domain(self):
        """Return the domain.

        ``f.domain`` is equivalent to ``f.get_domain()``

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get_domain`

        :Returns:

            `Domain`
                The domain.

        **Examples**

        >>> f = {{package}}.example_field(2)
        >>> f.domain
        <Domain: {1, 5, 8, 36}>

        >>> d0 = f.domain
        >>> d1 = f.get_domain()
        >>> d0.equals(d1)
        True

        """
        return self.get_domain()

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def del_data_axes(self, key=None, default=ValueError()):
        """Removes the keys of the axes spanned by the construct data.

        Specifically, removes the keys of the domain axis constructs
        spanned by the data of the field or of a metadata construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get_data_axes`, `has_data_axes`, `set_data_axes`

        :Parameters:

            key: `str`, optional
                Specify a metadata construct, instead of the field
                construct.

                *Parameter example:*
                  ``key='auxiliarycoordinate0'``

            default: optional
                Return the value of the *default* parameter if the data
                axes have not been set.

                {{default Exception}}

        :Returns:

            `tuple`
                The removed keys of the domain axis constructs spanned by
                the data.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> f.get_data_axes()
        ('domainaxis0', 'domainaxis1')
        >>> f.get_data_axes(key='dimensioncoordinate2')
        ('domainaxis2',)
        >>> f.has_data_axes()
        True

        >>> f.del_data_axes()
        ('domainaxis0', 'domainaxis1')
        >>> f.has_data_axes()
        False
        >>> f.get_data_axes(default='no axes')
        'no axes'

        """
        if key is not None:
            return super().del_data_axes(key, default=default)

        return self._del_component("data_axes", default=default)

    def get_domain(self):
        """Return the domain.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `domain`

        :Returns:

            `Domain`
                 The domain.

        **Examples**

        >>> f = {{package}}.example_field(2)
        >>> f.get_domain()
        <Domain: {1, 5, 8, 36}>

        >>> d0 = f.domain
        >>> d1 = f.get_domain()
        >>> d0.equals(d1)
        True

        """
        return self._Domain.fromconstructs(self.constructs, copy=False)

    def get_data_axes(self, key=None, default=ValueError()):
        """Gets the keys of the axes spanned by the construct data.

        Specifically, returns the keys of the domain axis constructs
        spanned by the data of the field or of a metadata construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_data_axes`, `get_data`, `set_data_axes`

        :Parameters:

            key: `str`, optional
                Specify a metadata construct, instead of the field
                construct.

                *Parameter example:*
                  ``key='auxiliarycoordinate0'``

            default: optional
                Return the value of the *default* parameter if the data
                axes have not been set.

                {{default Exception}}

        :Returns:

            `tuple`
                The keys of the domain axis constructs spanned by the
                data.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> f.get_data_axes()
        ('domainaxis0', 'domainaxis1')
        >>> f.get_data_axes(key='dimensioncoordinate2')
        ('domainaxis2',)
        >>> f.has_data_axes()
        True

        >>> f.del_data_axes()
        ('domainaxis0', 'domainaxis1')
        >>> f.has_data_axes()
        False
        >>> f.get_data_axes(default='no axes')
        'no axes'

        """
        if key is not None:
            return super().get_data_axes(key, default=default)

        return self._get_component("data_axes", default=default)

    def has_data_axes(self, key=None):
        """Whether the axes spanned by the construct data have been set.

        Specifically, whether the domain axis constructs spanned by the
        data of the field or of a metadata construct have been set.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_data_axes`, `get_data_axes`, `set_data_axes`

        :Parameters:

            key: `str`, optional
                Specify a metadata construct, instead of the field
                construct.

                *Parameter example:*
                  ``key='domainancillary1'``

        :Returns:

            `bool`
                True if domain axis constructs that span the data been
                set, otherwise False.

        **Examples**

        >>> f = {{package}}.example_field(0)
        >>> f.get_data_axes()
        ('domainaxis0', 'domainaxis1')
        >>> f.get_data_axes('dimensioncoordinate2')
        ('domainaxis2',)
        >>> f.has_data_axes()
        True

        >>> f.del_data_axes()
        ('domainaxis0', 'domainaxis1')
        >>> f.has_data_axes()
        False
        >>> f.get_data_axes(default='no axes')
        'no axes'

        """
        if key is None:
            axes = self.get_data_axes(default=None)
        else:
            axes = self.get_data_axes(key, default=None)

        if axes is None:
            return False

        return super().has_data_axes(key)

    def del_construct(self, key, default=ValueError()):
        """Remove a metadata construct.

        If a domain axis construct is selected for removal then it
        can't be spanned by any data arrays of the field nor metadata
        constructs, nor be referenced by any cell method
        constructs. However, a domain ancillary construct may be
        removed even if it is referenced by coordinate reference
        construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get_construct`, `constructs`, `has_construct`,
                     `set_construct`

        :Parameters:

            key: `str`
                The construct identifier of the metadata construct to
                be removed.

                *Parameter example:*
                  ``key='auxiliarycoordinate0'``

            default: optional
                Return the value of the *default* parameter if the
                data axes have not been set.

                {{default Exception}}

        :Returns:

                The removed metadata construct.

        **Examples**

        >>> f = {{package}}.example_field(4)
        >>> f.del_construct('auxiliarycoordinate2')
        <{{repr}}AuxiliaryCoordinate: longitude(3) degrees_east>

        >>> f = {{package}}.example_field(0)
        >>> f.del_construct('auxiliarycoordinate2')
        Traceback (most recent call last):
            ...
        ValueError: Can't remove non-existent construct
        >>> f.del_construct('auxiliarycoordinate2', default=False)
        False

        """
        domain_axes = self.constructs.filter_by_type(
            "domain_axis", todict=True
        )
        if key in domain_axes and key in self.get_data_axes(default=()):
            raise ValueError(
                f"Can't remove domain axis {key!r} that is spanned by the "
                "data of the field construct"
            )

        return self.constructs._del_construct(key, default=default)

    def set_data(self, data, axes=None, copy=True, inplace=True):
        """Set the data of the field construct.

        The units, calendar and fill value properties of the data
        object are removed prior to insertion.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_data`, `get_data`, `has_data`, `set_data_axes`

        :Parameters:

            data: data_like
                The data to be inserted.

                {{data_like}}

            axes: (sequence of) `str`, or `None`
                The identifiers of the domain axes spanned by the data
                array. If `None` then the data axes are not set.

                The axes may also be set afterwards with the
                `set_data_axes` method.

                *Parameter example:*
                  ``axes=['domainaxis2']``

                *Parameter example:*
                  ``axes='domainaxis2'``

                *Parameter example:*
                  ``axes=['domainaxis2', 'domainaxis1']``

                *Parameter example:*
                  ``axes=None``

            {{data copy: `bool`, optional}}

            {{inplace: `bool`, optional (default True)}}

                .. versionadded:: (cfdm) 1.8.7.0

        :Returns:

            `None` or `{{class}}`
                If the operation was in-place then `None` is returned,
                otherwise return a new `{{class}}` instance containing the
                new data.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_data([1, 2, 3])
        >>> f.has_data()
        True
        >>> f.get_data()
        <{{repr}}Data(3): [1, 2, 3]>
        >>> f.data
        <{{repr}}Data(3): [1, 2, 3]>

        >>> f.del_data()
        <{{repr}}Data(3): [1, 2, 3]>
        >>> g = f.set_data([4, 5, 6], inplace=False)
        >>> g.data
        <{{repr}}Data(3): [4, 5, 6]>
        >>> f.has_data()
        False
        >>> print(f.get_data(None))
        None
        >>> print(f.del_data(None))
        None

        """
        if inplace:
            f = self
        else:
            f = self.copy(data=False)

        if axes is None:
            existing_axes = f.get_data_axes(default=None)
            if existing_axes is not None:
                f.set_data_axes(axes=existing_axes, _shape=np.shape(data))
        else:
            f.set_data_axes(axes=axes, _shape=np.shape(data))

        super(Field, f).set_data(data, copy=copy, inplace=True)

        if inplace:
            return

        return f

    def set_data_axes(self, axes, key=None, _shape=None):
        """Sets the axes spanned by the construct data.

        Specifically, sets the domain axis constructs spanned by the
        data of the field or of a metadata construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_data_axes`, `get_data`, `get_data_axes`,
                     `has_data_axes`

        :Parameters:

             axes: sequence of `str`
                The identifiers of the domain axis constructs spanned by
                the data of the field or of a metadata construct.

                *Parameter example:*
                  ``axes='domainaxis1'``

                *Parameter example:*
                  ``axes=['domainaxis1']``

                *Parameter example:*
                  ``axes=['domainaxis1', 'domainaxis0']``

             key: `str`, optional
                Specify a metadata construct, instead of the field
                construct.

                *Parameter example:*
                  ``key='domainancillary1'``

        :Returns:

            `None`

        **Examples**

        Set the domain axis constructs spanned by the data of the field
        construct:

        >>> f = {{package}}.{{class}}()
        >>> f.set_data_axes(['domainaxis0', 'domainaxis1'])
        >>> f.get_data_axes()
        ('domainaxis0', 'domainaxis1')

        Set the domain axis constructs spanned by the data of a metadata
        construct:

        >>> f = {{package}}.example_field(5)
        >>> f.set_data_axes(['domainaxis1'], key='dimensioncoordinate1')

        """
        if isinstance(axes, str):
            axes = (axes,)

        if key is not None:
            return super().set_data_axes(axes=axes, key=key)

        if _shape is None:
            data = self.get_data(None)
            if data is not None:
                _shape = data.shape

        if _shape is not None:
            domain_axes = self.constructs.filter_by_type(
                "domain_axis", todict=True
            )
            axes_shape = []
            for axis in axes:
                if axis not in domain_axes:
                    raise ValueError(
                        "Can't set field construct data axes: Domain axis "
                        f"{axis!r} doesn't exist"
                    )

                axes_shape.append(domain_axes[axis].get_size())

            if _shape != tuple(axes_shape):
                raise ValueError(
                    "Can't set field construct data axes: Data array shape "
                    f"of {_shape!r} does not match the shape of the given "
                    f"domain axes {tuple(axes)}: {tuple(axes_shape)}"
                )

        axes = tuple(axes)
        self._set_component("data_axes", axes, copy=False)

        self.constructs._field_data_axes = axes



from ..functions import deepcopy
from .container import Container


class Parameters(Container):
    """Mixin class for a collection of named parameters.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(self, parameters=None, source=None, copy=True):
        """**Initialisation**

        :Parameters:

            parameters: `dict`, optional
               Set parameters. The dictionary keys are parameter
               names, with corresponding values.

               Parameters may also be set after initialisation with
               the `set_parameters` and `set_parameter` methods.

               *Parameter example:*
                 ``parameters={'earth_radius': 6371007.}``

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(source=source, copy=copy)

        self._set_component("parameters", {}, copy=False)

        if source:
            try:
                parameters = source.parameters()
            except AttributeError:
                parameters = None

        if parameters is None:
            parameters = {}
            copy = False

        self.set_parameters(parameters, copy=copy)

    def clear_parameters(self):
        """Remove all parameters.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_parameter`, `parameters`, `set_parameters`

        :Returns:

            `dict`
                The parameters that have been removed.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.parameters()
        {}
        >>> p = {'standard_parallel': 25.0,
        ...      'longitude_of_central_meridian': 265.0,
        ...      'latitude_of_projection_origin': 25.0}
        >>> f.set_parameters(p)
        >>> f.parameters()
        {'standard_parallel': 25.0,
         'longitude_of_central_meridian': 265.0,
         'latitude_of_projection_origin': 25.0}

        >>> old = f.clear_parameters()
        >>> old
        {'standard_parallel': 25.0,
         'longitude_of_central_meridian': 265.0,
         'latitude_of_projection_origin': 25.0}
        >>> f.set_parameters(old)

        """
        out = self._get_component("parameters")
        self._set_component("parameters", {}, copy=False)
        return out.copy()

    def del_parameter(self, parameter, default=ValueError()):
        """Delete a parameter.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get_parameter`, `has_parameter`, `parameters`,
                     `set_parameter`

        :Parameters:

            parameter: `str`
                The name of the parameter to be deleted.

            default: optional
                Return the value of the *default* parameter if the
                parameter has not been set.

                {{default Exception}}

        :Returns:

                The removed parameter value.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_parameter('earth_radius', 6371007)
        >>> f.has_parameter('earth_radius')
        True
        >>> f.get_parameter('earth_radius')
        6371007
        >>> f.del_parameter('earth_radius')
        6371007
        >>> f.has_parameter('earth_radius')
        False
        >>> print(f.del_parameter('earth_radius', None))
        None
        >>> print(f.get_parameter('earth_radius', None))
        None

        """
        try:
            return self._get_component("parameters").pop(parameter)
        except KeyError:
            if default is None:
                return

            return self._default(
                default,
                f"{self.__class__.__name__!r} has no {parameter!r} parameter",
            )

    def get_parameter(self, parameter, default=ValueError()):
        """Get a parameter value.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            parameter: `str`
                The name of the parameter.

            default: optional
                Return the value of the *default* parameter if the
                parameter has not been set.

                {{default Exception}}

        :Returns:

                The value of the parameter.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_parameter('earth_radius', 6371007)
        >>> f.has_parameter('earth_radius')
        True
        >>> f.get_parameter('earth_radius')
        6371007
        >>> f.del_parameter('earth_radius')
        6371007
        >>> f.has_parameter('earth_radius')
        False
        >>> print(f.del_parameter('earth_radius', None))
        None
        >>> print(f.get_parameter('earth_radius', None))
        None

        """
        try:
            return self._get_component("parameters")[parameter]
        except KeyError:
            if default is None:
                return

            return self._default(
                default,
                f"{self.__class__.__name__!r} has no {parameter!r} parameter",
            )

    def has_parameter(self, parameter):
        """Whether a parameter has been set.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_parameter`, `get_parameter`, `parameters`,
                     `set_parameter`

        :Parameters:

            parameter: `str`
                The name of the parameter.

                *Parameter example:*
                   ``parameter='geoid_name'``

        :Returns:

             `bool`
                True if the parameter has been set, otherwise False.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_parameter('earth_radius', 6371007)
        >>> f.has_parameter('earth_radius')
        True
        >>> f.get_parameter('earth_radius')
        6371007
        >>> f.del_parameter('earth_radius')
        6371007
        >>> f.has_parameter('earth_radius')
        False
        >>> print(f.del_parameter('earth_radius', None))
        None
        >>> print(f.get_parameter('earth_radius', None))
        None

        """
        return parameter in self._get_component("parameters")

    def parameters(self):
        """Return all parameters.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `clear_parameters`, `get_parameter`, `has_parameter`
                     `set_parameters`

        :Returns:

            `dict`
                The parameters.

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.parameters()
        {}
        >>> p = {'standard_parallel': 25.0,
        ...      'longitude_of_central_meridian': 265.0,
        ...      'latitude_of_projection_origin': 25.0}
        >>> f.set_parameters(p)
        >>> f.parameters()
        {'standard_parallel': 25.0,
         'longitude_of_central_meridian': 265.0,
         'latitude_of_projection_origin': 25.0}

        >>> old = f.clear_parameters()
        >>> old
        {'standard_parallel': 25.0,
         'longitude_of_central_meridian': 265.0,
         'latitude_of_projection_origin': 25.0}
        >>> f.set_parameters(old)

        """
        return self._get_component("parameters").copy()

    def set_parameters(self, parameters, copy=True):
        """Set parameters.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `clear_parameters`, `parameters`, `set_parameter`

        :Parameters:

            parameters: `dict`
                Store the parameters from the dictionary supplied.

                *Parameter example:*
                  ``parameters={'earth_radius': 6371007}``

            copy: `bool`, optional
                If False then any parameter values provided by the
                *parameters* parameter are not copied before insertion. By
                default they are deep copied.

        :Returns:

            `None`

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.parameters()
        {}
        >>> p = {'standard_parallel': 25.0,
        ...      'longitude_of_central_meridian': 265.0,
        ...      'latitude_of_projection_origin': 25.0}
        >>> f.set_parameters(p)
        >>> f.parameters()
        {'standard_parallel': 25.0,
         'longitude_of_central_meridian': 265.0,
         'latitude_of_projection_origin': 25.0}

        >>> old = f.clear_parameters()
        >>> old
        {'standard_parallel': 25.0,
         'longitude_of_central_meridian': 265.0,
         'latitude_of_projection_origin': 25.0}
        >>> f.set_parameters(old)

        """
        if copy:
            parameters = deepcopy(parameters)

        self._get_component("parameters").update(parameters)

    def set_parameter(self, term, value, copy=True):
        """Set a parameter-valued term.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `parameters`

        :Returns:

            `None`

        **Examples**

        >>> f = {{package}}.{{class}}()
        >>> f.set_parameter('earth_radius', 6371007)
        >>> f.has_parameter('earth_radius')
        True
        >>> f.get_parameter('earth_radius')
        6371007
        >>> f.del_parameter('earth_radius')
        6371007
        >>> f.has_parameter('earth_radius')
        False
        >>> print(f.del_parameter('earth_radius', None))
        None
        >>> print(f.get_parameter('earth_radius', None))
        None

        """
        if copy:
            value = deepcopy(value)

        self._get_component("parameters")[term] = value



import atexit
import datetime
import faulthandler
import os
import tempfile
import unittest
from urllib.parse import urlparse

faulthandler.enable()  # to debug seg faults and timeouts

import numpy as np
from dask.base import tokenize

import cfdm

n_tmpfiles = 1
tmpfiles = [
    tempfile.mkstemp("_test_netCDF.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
(tmpfile,) = tmpfiles


def _remove_tmpfiles():
    """Remove temporary files created during tests."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)


class NetCDF4ArrayTest(unittest.TestCase):
    """Unit test for the NetCDF4Array class."""

    f0 = cfdm.example_field(0)

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

    def test_NetCDF4Array_get_address(self):
        """Test NetCDF4Array.get_address."""
        a = cfdm.NetCDF4Array(address="tas")
        self.assertEqual(a.get_address(), "tas")

        a = cfdm.NetCDF4Array()
        self.assertIsNone(a.get_address(default=None))

    def test_NetCDF4Array_get_filename(self):
        """Test NetCDF4Array.get_filename."""
        a = cfdm.NetCDF4Array("/data1/file1")
        self.assertEqual(a.get_filename(), "/data1/file1")

        a = cfdm.NetCDF4Array()
        self.assertIsNone(a.get_filename(default=None))

    def test_NetCDF4Array_mask(self):
        """Test NetCDF4Array masking."""
        f = self.f0
        f.data[0] = np.ma.masked
        cfdm.write(f, tmpfile)
        array = f.array

        n = cfdm.NetCDF4Array(tmpfile, f.nc_get_variable(), shape=f.shape)
        self.assertTrue(n.get_mask())
        n = np.asanyarray(n[...])
        self.assertTrue((array.mask == n.mask).all())

        n = cfdm.NetCDF4Array(
            tmpfile, f.nc_get_variable(), shape=f.shape, mask=False
        )
        self.assertFalse(n.get_mask())
        n = np.asanyarray(n[...])
        self.assertEqual(np.ma.count(n), n.size)

    def test_NetCDF4Array_unpack(self):
        """Test NetCDF4Array unpacking."""
        add_offset = 10.0
        scale_factor = 3.14

        f = self.f0.copy()
        f.data[0] = np.ma.masked
        array0 = f.array
        array1 = (array0 - add_offset) / scale_factor

        f.set_property("add_offset", add_offset)
        f.set_property("scale_factor", scale_factor)
        cfdm.write(f, tmpfile)

        n = cfdm.NetCDF4Array(tmpfile, f.nc_get_variable(), shape=f.shape)
        self.assertTrue(n.get_unpack())
        n = np.asanyarray(n[...])
        self.assertTrue((n.mask == array0.mask).all())
        self.assertTrue(np.ma.allclose(n, array0))

        n = cfdm.NetCDF4Array(
            tmpfile, f.nc_get_variable(), shape=f.shape, unpack=False
        )
        self.assertFalse(n.get_unpack())
        n = np.asanyarray(n[...])
        self.assertTrue((n.mask == array1.mask).all())
        self.assertTrue((n == array1).all())

    def test_NetCDF4Array_get_storage_options(self):
        """Test NetCDF4Array get_storage_options."""
        n = cfdm.NetCDF4Array(filename="filename.nc")
        self.assertEqual(n.get_storage_options(), {})

        n = cfdm.NetCDF4Array(
            filename="filename.nc", storage_options={"anon": True}
        )
        self.assertEqual(n.get_storage_options(), {"anon": True})

        n = cfdm.NetCDF4Array(filename="s3://store/filename.nc")
        self.assertEqual(
            n.get_storage_options(), {"endpoint_url": "https://store"}
        )
        self.assertEqual(n.get_storage_options(create_endpoint_url=False), {})

        n = cfdm.NetCDF4Array(
            filename="s3://store/filename.nc", storage_options={"anon": True}
        )
        self.assertEqual(
            n.get_storage_options(),
            {"anon": True, "endpoint_url": "https://store"},
        )
        self.assertEqual(
            n.get_storage_options(create_endpoint_url=False), {"anon": True}
        )
        other_file = "s3://other/file.nc"
        self.assertEqual(
            n.get_storage_options(filename=other_file),
            {"anon": True, "endpoint_url": "https://other"},
        )
        self.assertEqual(
            n.get_storage_options(parsed_filename=urlparse(other_file)),
            {"anon": True, "endpoint_url": "https://other"},
        )

        n = cfdm.NetCDF4Array(
            filename="s3://store/filename.nc",
            storage_options={"anon": True, "endpoint_url": None},
        )
        self.assertEqual(
            n.get_storage_options(),
            {"anon": True, "endpoint_url": None},
        )

    def test_NetCDF4Array_get_attributes(self):
        """Test NetCDF4Array get_attributes."""
        f = self.f0
        cfdm.write(f, tmpfile)
        n = cfdm.NetCDF4Array(tmpfile, f.nc_get_variable(), shape=f.shape)
        self.assertEqual(n.get_attributes(), {})

        # Set attributes via indexing
        n = n[...]
        _ = np.asanyarray(n)
        self.assertEqual(
            n.get_attributes(),
            {
                "cell_methods": "area: mean",
                "coordinates": "time",
                "project": "research",
                "standard_name": "specific_humidity",
                "units": "1",
            },
        )

    def test_NetCDF4Array_file_directory(self):
        """Test NetCDF4Array.file_directory."""
        a = cfdm.NetCDF4Array("/data1/file1")
        self.assertEqual(a.file_directory(), "/data1")

        a = cfdm.NetCDF4Array()
        self.assertIsNone(a.file_directory(default=None))

    def test_NetCDF4Array__dask_tokenize__(self):
        """Test NetCDF4Array.__dask_tokenize__"""
        a = cfdm.NetCDF4Array("/data1/file1", "tas", shape=(12, 2), mask=False)
        self.assertEqual(tokenize(a), tokenize(a.copy()))

        b = cfdm.NetCDF4Array("/home/file2", "tas", shape=(12, 2))
        self.assertNotEqual(tokenize(a), tokenize(b))

    def test_NetCDF4Array_shape(self):
        """Test NetCDF4Array.shape."""
        shape = (12, 73, 96)
        a = cfdm.NetCDF4Array("/home/file2", "tas", shape=shape)
        self.assertEqual(a.shape, shape)
        self.assertEqual(a.original_shape, shape)
        a = a[::2]
        self.assertEqual(a.shape, (shape[0] // 2,) + shape[1:])
        self.assertEqual(a.original_shape, shape)

    def test_NetCDF4Array_index(self):
        """Test NetCDF4Array.index."""
        shape = (12, 73, 96)
        a = cfdm.NetCDF4Array("/home/file2", "tas", shape=shape)
        self.assertEqual(list(a.index()), [slice(0, n, 1) for n in shape])
        a = a[8:7:-1, 10:19:3, [15, 1, 4, 12]]
        a = a[[0], [True, False, True], ::-2]
        self.assertEqual(a.shape, (1, 2, 2))
        self.assertEqual(
            a.index(),
            (slice(8, 9, None), slice(10, 17, 6), slice(12, -1, -11)),
        )

        index = a.index(conform=False)
        self.assertTrue((index[0] == [8]).all())
        self.assertTrue((index[1] == [10, 16]).all())
        self.assertTrue((index[2] == [12, 1]).all())

        # New dimensions
        a = cfdm.NetCDF4Array("/home/file2", "tas", shape=shape)

        b = a[:2, None, ...]
        self.assertEqual(
            b.index(), (slice(0, 2, 1), None, slice(0, 73, 1), slice(0, 96, 1))
        )
        self.assertEqual(b.shape, (2, 1, 73, 96))
        self.assertEqual(b.reference_shape, (12, 1, 73, 96))

        b = b[...]
        self.assertEqual(
            b.index(), (slice(0, 2, 1), None, slice(0, 73, 1), slice(0, 96, 1))
        )
        self.assertEqual(b.shape, (2, 1, 73, 96))
        self.assertEqual(b.reference_shape, (12, 1, 73, 96))

        b = b[..., :4]
        self.assertEqual(
            b.index(), (slice(0, 2, 1), None, slice(0, 73, 1), slice(0, 4, 1))
        )
        self.assertEqual(b.shape, (2, 1, 73, 4))
        self.assertEqual(b.reference_shape, (12, 1, 73, 96))

        b = b[..., None, :]
        self.assertEqual(
            b.index(),
            (slice(0, 2, 1), None, slice(0, 73, 1), None, slice(0, 4, 1)),
        )
        self.assertEqual(b.shape, (2, 1, 73, 1, 4))
        self.assertEqual(b.reference_shape, (12, 1, 73, 1, 96))

        b = b[..., None]
        self.assertEqual(
            b.index(),
            (
                slice(0, 2, 1),
                None,
                slice(0, 73, 1),
                None,
                slice(0, 4, 1),
                None,
            ),
        )
        self.assertEqual(b.shape, (2, 1, 73, 1, 4, 1))
        self.assertEqual(b.reference_shape, (12, 1, 73, 1, 96, 1))

        with self.assertRaises(IndexError):
            # index 4 is out of bounds for axis 5 with size 1
            b[..., 4]

        b = b[:, 0, :, -1, :, 0]
        self.assertEqual(
            b.index(), (slice(0, 2, 1), slice(0, 73, 1), slice(0, 4, 1))
        )
        self.assertEqual(b.shape, (2, 73, 4))
        self.assertEqual(b.original_shape, shape)
        self.assertEqual(b.reference_shape, shape)

    def test_NetCDF4Array_replace_directory(self):
        """Test NetCDF4Array.replace_directory."""
        cwd = os.getcwd()

        n = cfdm.NetCDF4Array("basename.nc")

        m = n.replace_directory()
        self.assertEqual(m.get_filename(), "basename.nc")
        m = n.replace_directory(new="data")
        self.assertEqual(m.get_filename(), "data/basename.nc")
        m = n.replace_directory(normalise=True)
        self.assertEqual(m.get_filename(), os.path.join(cwd, "basename.nc"))

        n = cfdm.NetCDF4Array("data/basename.nc")

        m = n.replace_directory()
        self.assertEqual(m.get_filename(), "data/basename.nc")
        m = n.replace_directory(new="/home")
        self.assertEqual(m.get_filename(), "/home/data/basename.nc")
        m = n.replace_directory(old="data")
        self.assertEqual(m.get_filename(), "basename.nc")
        m = n.replace_directory(old="data", new="home")
        self.assertEqual(m.get_filename(), "home/basename.nc")
        m = n.replace_directory(normalise=True)
        self.assertEqual(
            m.get_filename(), os.path.join(cwd, "data/basename.nc")
        )
        m = n.replace_directory(old=cwd, new="new", normalise=True)
        self.assertEqual(
            m.get_filename(), os.path.join(cwd, "new/data/basename.nc")
        )

        n = cfdm.NetCDF4Array("path/basename.nc")
        m = n.replace_directory("path", "../new_path")
        self.assertEqual(m.get_filename(), "../new_path/basename.nc")

        n = cfdm.NetCDF4Array("/data/../new_path/basename.nc")
        m = n.replace_directory("/new_path/", normalise=True)
        self.assertEqual(m.get_filename(), "basename.nc")

        n = cfdm.NetCDF4Array("/data/basename.nc")
        m = n.replace_directory(new="/home")
        self.assertEqual(m.get_filename(), "/home/data/basename.nc")

        n = cfdm.NetCDF4Array("/data/basename.nc")
        m = n.replace_directory(new="/home/")
        self.assertEqual(m.get_filename(), "/home/data/basename.nc")
        m = n.replace_directory(old="/data")
        self.assertEqual(m.get_filename(), "basename.nc")

    def test_NetCDF4Array_concatenate(self):
        """Test np.concatenate on NetCDF4Array objects."""
        f = self.f0
        f.data[0] = np.ma.masked
        cfdm.write(f, tmpfile)

        array = f.array
        c0 = np.ma.concatenate((array, array), axis=0)

        n = cfdm.NetCDF4Array(
            tmpfile, f.nc_get_variable(), shape=f.shape, dtype=f.dtype
        )

        c1 = np.concatenate((n, n), axis=0)
        self.assertTrue(np.ma.is_masked(c1))

        self.assertEqual(c0.shape, c1.shape)
        self.assertTrue((c0.mask == c1.mask).all())
        self.assertTrue((c0 == c1).all())

    def test_NetCDF4Array_astype(self):
        """Test NetCDF4Array.astype."""
        f = self.f0
        cfdm.write(f, tmpfile)

        n = cfdm.NetCDF4Array(
            tmpfile, f.nc_get_variable(), shape=f.shape, dtype=f.dtype
        )
        self.assertNotEqual(n.dtype, np.dtype("int32"))

        a = n.astype("int32")
        self.assertIsInstance(a, np.ndarray)
        self.assertEqual(a.dtype, np.dtype("int32"))
        self.assertTrue((a == f.array.astype("int32")).all())


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



import numpy as np

from ..netcdfindexer import netcdf_indexer
from .array import Array


class DeprecationError(Exception):
    """Deprecation error."""

    pass


class CompressedArray(Array):
    """Mixin class for a container of an underlying compressed array.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __new__(cls, *args, **kwargs):
        """Store subarray classes.

        A child class must define its subarray classes in the
        `_Subarray` dictionary.

        .. versionadded:: (cfdm) 1.10.0.0

        """
        instance = super().__new__(cls)
        instance._Subarray = {}
        return instance

    def __init__(
        self,
        compressed_array=None,
        shape=None,
        size=None,
        ndim=None,
        compressed_dimension=None,
        compressed_dimensions=None,
        compression_type=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            compressed_array: array_like
                The compressed array.

            shape: `tuple`
                The uncompressed array dimension sizes.

            compressed_dimensions: `dict`
                Mapping of dimensions of the compressed array to their
                corresponding dimensions in the uncompressed
                array. Compressed array dimensions that are not
                compressed must be omitted from the mapping.

                *Parameter example:*
                  ``{0: (0, 1)}``

                *Parameter example:*
                  ``{0: (0, 1, 2)}``

                *Parameter example:*
                  ``{2: (2, 3)}``

                *Parameter example:*
                  ``{1: (1,)}``

                *Parameter example:*
                  ``{0: (0,), 2: (2,)}``

                .. versionadded:: (cfdm) 1.10.0.0

            compression_type: `str`
                The type of compression.

            kwargs: optional
                Further named parameters and their values needed to define
                the compressed array.

            compressed_dimension: Deprecated at version 1.10.0.0
                Use the *compressed_dimensions* parameter instead.

            {{init source: optional}}

                .. versionadded:: (cfdm) 1.10.0.0

            copy: `bool`, optional
                If True (the default) then deep copy input parameters
                prior to initialisation. If False then arguments are
                not deep copied.

                .. versionadded:: (cfdm) 1.10.0.0

            size: `int`
                Deprecated at version 1.10.0.0. If set will be
                ignored.

                Number of elements in the uncompressed array.

            ndim: `int`
                Deprecated at version 1.10.0.0. If set will be
                ignored.

                The number of uncompressed array dimensions

        """
        if compressed_dimension is not None:
            raise DeprecationError(
                "The 'compressed_dimension' keyword was deprecated at "
                "version 1.10.0.0. "
                "Use the 'compressed_dimensions' keyword instead."
            )  # pragma: no cover

        super().__init__(source=source, copy=copy)

        if source is not None:
            try:
                compressed_array = source._get_compressed_Array(None)
            except AttributeError:
                compressed_array = None

            try:
                compression_type = source.get_compression_type()
            except AttributeError:
                compression_type = None

            try:
                compressed_dimensions = source.compressed_dimensions()
            except AttributeError:
                compressed_dimensions = {}

            try:
                shape = source.shape
            except AttributeError:
                shape = None

        elif compressed_dimensions is None:
            compressed_dimensions = {}
        else:
            compressed_dimensions = compressed_dimensions.copy()

        self._set_component(
            "compressed_dimensions", compressed_dimensions, copy=False
        )

        if compressed_array is not None:
            self._set_compressed_Array(compressed_array, copy=copy)

        if shape is not None:
            self._set_component("shape", shape, copy=False)

        if compression_type:
            self._set_component(
                "compression_type", compression_type, copy=False
            )

    def __getitem__(self, indices):
        """Return a subspace of the uncompressed data.

        x.__getitem__(indices) <==> x[indices]

        Returns a subspace of the uncompressed array as an independent
        numpy array.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        # ------------------------------------------------------------
        # Method: Uncompress the entire array and then subspace it
        # ------------------------------------------------------------
        # Initialise the un-sliced uncompressed array
        u = np.ma.masked_all(self.shape, dtype=self.dtype)

        Subarray = self.get_Subarray()
        subarray_kwargs = {
            **self.conformed_data(),
            **self.subarray_parameters(),
        }

        for u_indices, u_shape, c_indices, _ in zip(*self.subarrays()):
            subarray = Subarray(
                indices=c_indices,
                shape=u_shape,
                **subarray_kwargs,
            )
            u[u_indices] = subarray[...]

        u = netcdf_indexer(
            u,
            mask=False,
            unpack=False,
            always_masked_array=False,
            orthogonal_indexing=True,
            copy=False,
        )
        return u[indices]

    def _first_or_last_element(self, indices):
        """Return the first or last element of the compressed array.

        This method will return the first or last element of the
        compressed array without performing any decompression.

        First and last elements are only recognised by exact *indices*
        matches to:

        * ``(slice(0, 1, 1),) * self.ndim``
        * ``(slice(-1, None, 1),) * self.ndim``

        Any other value of indices will raise an `IndexError`. For
        instance, note that ``slice(0, 1, 1)`` is not an exact match
        to ``slice(0, 1)``.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `__getitem__`

        :Parameters:

            indices:
                Indices to the uncompressed array.

        :Returns:

            `numpy.ndarray`
                The first or last element. If the *indices* do not
                acceptably select the first or last element then an
                `IndexError` is raised.

        """
        ndim = self.ndim
        for index in (slice(0, 1, 1), slice(-1, None, 1)):
            if indices == (index,) * ndim:
                data = self.source()
                return np.asanyarray(data[(index,) * data.ndim])

        # Indices do not acceptably select the first nor last element
        raise IndexError()

    def _get_compressed_Array(self, default=ValueError()):
        """Return the compressed array.

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the array
                has not been set. If set to an `Exception` instance then
                it will be raised instead.

        :Returns:

                The compressed Array instance.

        **Examples**

        >>> c = d._get_compressed_Array()

        """
        return self._get_component("compressed_Array", default)

    def _set_compressed_Array(self, array, copy=True):
        """Set the compressed array.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            array: subclass of `cfdm.data.Array`
                The compressed data to be inserted.

        :Returns:

            `None`

        **Examples**

        >>> d._set_compressed_Array(a)

        """
        if copy:
            array = array.copy()

        self._set_component("compressed_Array", array, copy=False)

    @property
    def array(self):
        """Returns a numpy array containing the uncompressed data.

        :Returns:

            `numpy.ndarray`
                The uncompressed array.

        **Examples**

        >>> n = a.array
        >>> isinstance(n, numpy.ndarray)
        True

        """
        return self[...]

    @property
    def dtype(self):
        """Data-type of the uncompressed data."""
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.dtype"
        )  # pragma: no cover

    @property
    def shape(self):
        """Shape of the uncompressed data."""
        return self._get_component("shape")

    @property
    def compressed_array(self):
        """Returns an independent numpy array with the compressed data.

        :Returns:

            `numpy.ndarray`
                The compressed array.

        **Examples**

        >>> n = a.compressed_array

        >>> isinstance(n, numpy.ndarray)
        True

        """
        ca = self._get_compressed_Array(None)
        if ca is None:
            raise ValueError("There is no underlying compressed array")

        try:
            return ca.array
        except AttributeError:
            return ca

    def get_compressed_axes(self):
        """Return axes that are compressed in the underlying array.

        :Returns:

            `list`
                The compressed axes described by their integer
                positions in the uncompressed array.

        **Examples**

        >>> c.ndim
        4
        >>> c.compressed_array.ndim
        3
        >>> c.compressed_axes()
        [1, 2]

        """
        return sorted(
            {x for y in self.compressed_dimensions().values() for x in y}
        )

    def get_compressed_dimension(self, default=ValueError()):
        """Returns the compressed dimension's position in the array.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get_compressed_axes`, `get_compression_type`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                underlying array is not compressed. If set to an
                `Exception` instance then it will be raised instead.

        :Returns:

            `int`
                The position of the compressed dimension in the compressed
                array. If the underlying is not compressed then *default*
                is returned, if provided.

        **Examples**

        >>> i = d.get_compressed_dimension()

        """
        compressed_dimensions = self.compressed_dimensions()
        if len(compressed_dimensions) != 1:
            raise ValueError(
                "Can't get unique compressed dimension when there "
                f"is not exactly one of them: {self.compressed_dimensions()}"
            )

        return tuple(compressed_dimensions)[0]

    def compressed_dimensions(self):
        """Mapping of compressed to uncompressed dimensions.

        A dictionary key is a position of a dimension in the
        compressed data, with a value of the positions of the
        corresponding dimensions in the uncompressed data. Compressed
        array dimensions that are not compressed are omitted from the
        mapping.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `get_compressed_axes`, `get_compression_type`

        :Returns:

            `dict`
                The mapping of dimensions of the compressed array to
                their corresponding dimensions in the uncompressed
                array. Compressed array dimensions that are not
                compressed are omitted from the mapping.

        """
        c = self._get_component("compressed_dimensions")
        if not c:
            raise ValueError("No compressed dimensions have been defined")

        return self._get_component("compressed_dimensions").copy()

    def conformed_data(self):
        """The data as required by the decompression algorithm.

        .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            `dict`
                The compressed data, with the key ``'data'``.

        """
        return {"data": self.source().copy()}

    def get_filename(self, normalise=False, default=AttributeError()):
        """Return the name of the file containing the compressed data.

        .. versionadded:: (cfdm) 1.10.0.2

        :Parameters:

            {{normalise: `bool`, optional}}

                .. versionadded:: (cfdm) 1.12.0.0

            default: optional
                Return the value of the *default* parameter if there
                is no file name.

                {{default Exception}}

                .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

            `str`
                The file name.

        """
        data = self._get_compressed_Array(None)
        if data is None:
            return set()

        filenames = data.get_filenames(normalise=normalise)
        if len(filenames) != 1:
            if default is None:
                return

            return self._default(
                default, f"{self.__class__.__name__} has no unique file name"
            )

        return filenames[0]

    def get_Subarray(self):
        """Return the Subarray class.

        .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            `Subarray`
                The class for representing subarrays.

        """
        return self._Subarray[self.get_compression_type()]

    def source(self, default=ValueError()):
        """Return the underlying array object.

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the array
                has not been set. If set to an `Exception` instance then
                it will be raised instead.

        :Returns:

            subclass of `Array`
                The underlying array object.

        **Examples**

        >>> array
        <RaggedContiguousArray(2, 4): >
        >>> array.source()
        <Data(5): [280.0, ..., 279.5]>

        """
        return self._get_compressed_Array(default=default)

    def subarray_parameters(self):
        """Non-data parameters required by the `Subarray` class.

        .. versionadded:: (cfdm) 1.11.0.0

        :Returns:

            `dict`
                The parameters as a dictionary value, with key
                ``'compressed_dimensions'``.

        """
        return {"compressed_dimensions": self.compressed_dimensions()}

    def subarray_shapes(self, shapes):
        """Create the subarray shapes along each uncompressed dimension.

        .. versionadded:: (cfdm) 1.10.0.0

        .. seealso:: `subarray`

        :Parameters:

            {{subarray_shapes chunks: `int`, sequence, `dict`, or `str`, optional}}

        :Returns:

            `list`
                 The subarray shapes along each uncompressed
                 dimension.

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.subarray_shapes"
        )  # pragma: no cover

    def subarrays(self, shapes=-1):
        """Return descriptors for every subarray.

        Theses descriptors are used during subarray decompression.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            {{subarrays chunks: ``-1`` or sequence, optional}}

        :Returns:

            sequence of iterable
                Each iterable iterates over a particular descriptor
                from each subarray.

        """
        raise NotImplementedError(
            f"Must implement {self.__class__.__name__}.subarrays"
        )  # pragma: no cover

    def to_memory(self):
        """Bring data on disk into memory.

        There is no change to data that is already in memory.

        .. versionadded:: (cfdm) 1.10.0.0

        :Returns:

            `{{class}}`
                A copy of the array with all of its data in memory.

        """
        a = self.copy()
        try:
            a._set_compressed_Array(a.source().to_memory(), copy=False)
        except AttributeError:
            pass

        return a



from ..meta import DocstringRewriteMeta


class FieldDomain(metaclass=DocstringRewriteMeta):
    """Mixin class for methods of field and domain constructs.

    .. versionadded:: (cfdm) 1.8.9.0

    """

    def __docstring_package_depth__(self):
        """Returns the package depth for {{package}} substitutions.

        See `_docstring_package_depth` for details.

        """
        return 1

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def del_construct(self, key, default=ValueError()):
        """Remove a metadata construct.

        If a domain axis construct is selected for removal then it
        can't be spanned by any metadata construct data, nor the field
        construct's data; nor be referenced by any cell method
        constructs.

        However, a domain ancillary construct may be removed even if
        it is referenced by coordinate reference construct. In this
        case the reference is replace with `None`.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `constructs`, `get_construct`, `has_construct`,
                     `set_construct`
        :Parameters:

            key: `str`
                The key of the metadata construct to be removed.
                *Parameter example:*
                  ``key='auxiliarycoordinate0'``

            default: optional
                Return the value of the *default* parameter if the
                construct can not be removed, or does not exist.
                {{default Exception}}

        :Returns:

                The removed metadata construct.

        **Examples**

        >>> f.del_construct('dimensioncoordinate1')
        <{{repr}}DimensionCoordinate: grid_latitude(111) degrees>

        """
        return self.constructs._del_construct(key, default=default)

    def get_construct(self, key, default=ValueError()):
        """Return a metadata construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `constructs`, `del_construct`, `has_construct`,
                     `set_construct`

        :Parameters:

            key: `str`
                The key of the metadata construct.

                *Parameter example:*
                  ``key='domainaxis1'``

            default: optional
                Return the value of the *default* parameter if the
                construct does not exist.

                {{default Exception}}

        :Returns:

                The metadata construct.

        **Examples**

        >>> f.constructs()
        {'auxiliarycoordinate0': <{{repr}}AuxiliaryCoordinate: latitude(10, 9) degree_N>,
         'auxiliarycoordinate1': <{{repr}}AuxiliaryCoordinate: longitude(9, 10) degreeE>,
         'auxiliarycoordinate2': <{{repr}}AuxiliaryCoordinate: long_name:greek_letters(10) >,
         'coordinatereference1': <{{repr}}CoordinateReference: rotated_latitude_longitude>,
         'dimensioncoordinate1': <{{repr}}DimensionCoordinate: grid_latitude(10) degrees>,
         'dimensioncoordinate2': <{{repr}}DimensionCoordinate: grid_longitude(9) degrees>,
         'domainaxis1': <{{repr}}DomainAxis: 10>,
         'domainaxis2': <{{repr}}DomainAxis: 9>}
        >>> f.get_construct('dimensioncoordinate1')
        <{{repr}}DimensionCoordinate: grid_latitude(10) degrees>

        """
        construct = self.constructs.get(key)
        if construct is not None:
            return construct

        if default is None:
            return default

        return self._default(default, f"No construct for key {key!r}")

    def has_construct(self, key):
        """Whether a metadata construct exists.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `constructs`, `del_construct`, `get_construct`,
                     `set_construct`

        :Parameters:

            key: `str`
                The identifier of the metadata construct.

                *Parameter example:*
                  ``key='auxiliarycoordinate0'``

        :Returns:

            `bool`
                True if the construct exists, otherwise False.

        **Examples**

        >>> f.has_construct('dimensioncoordinate1')
        True

        >>> f.has_construct('dimensioncoordinate99')
        False

        """
        return key in self.constructs

    def set_construct(self, construct, key=None, axes=None, copy=True):
        """Set a metadata construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `constructs`, `del_construct`, `get_construct`,
                     `set_data_axes`

        :Parameters:

            construct:
                The metadata construct to be inserted.

            key: `str`, optional
                The construct identifier to be used for the
                construct. If not set then a new, unique identifier is
                created automatically. If the identifier already
                exists then the existing construct will be replaced.

                *Parameter example:*
                  ``key='cellmeasure0'``

            axes: (sequence of) `str`, optional
                The construct identifiers of the domain axis
                constructs spanned by the data array. An exception is
                raised if used for a metadata construct that can not
                have a data array, i.e. domain axis, cell method and
                coordinate reference constructs.

                *Parameter example:*
                  ``axes='domainaxis1'``

                *Parameter example:*
                  ``axes=['domainaxis1']``

                *Parameter example:*
                  ``axes=['domainaxis1', 'domainaxis0']``

            copy: `bool`, optional
                If True then set a copy of the construct. By default
                the construct is copied.

        :Returns:

             `str`
                The construct identifier for the construct.

        **Examples**

        >>> key = f.set_construct(c)
        >>> key = f.set_construct(c, copy=False)
        >>> key = f.set_construct(c, axes='domainaxis2')
        >>> key = f.set_construct(c, key='cellmeasure0')

        """
        return self.constructs._set_construct(
            construct, key=key, axes=axes, copy=copy
        )

    def get_data_axes(self, key, default=ValueError):
        """Gets the keys of the axes spanned by a construct's data.

        Specifically, returns the keys of the domain axis constructs
        spanned by the data of a metadata construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_data_axes`, `has_data_axes`, `set_data_axes`

        :Parameters:

            key: `str`
                Specify a metadata construct.

                *Parameter example:*
                  ``key='auxiliarycoordinate0'``

            default: optional
                Return the value of the *default* parameter if the
                data axes have not been set.

                {{default Exception}}

        :Returns:

            `tuple`
                The keys of the domain axis constructs spanned by the
                data.

        **Examples**

        >>> f.set_data_axes(['domainaxis0', 'domainaxis1'])
        >>> f.get_data_axes()
        ('domainaxis0', 'domainaxis1')
        >>> f.del_data_axes()
        ('domainaxis0', 'domainaxis1')
        >>> print(f.del_dataxes(None))
        None
        >>> print(f.get_data_axes(default=None))
        None

        """
        data_axes = self.constructs.get_data_axes(key, default=None)
        if data_axes is not None:
            return data_axes

        if default is None:
            return default

        return self._default(
            default,
            message=(
                f"{self.__class__.__name__!r} has no data axes for "
                f"construct {key!r}"
            ),
        )

    def del_data_axes(self, key, default=ValueError()):
        """Removes the keys of the axes spanned by the construct data.

        Specifically, removes the keys of the domain axis constructs
        spanned by the data of a metadata construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `get_data_axes`, `has_data_axes`, `set_data_axes`

        :Parameters:

            key: `str`
                Specify a metadata construct, instead of the field
                construct.

                *Parameter example:*
                  ``key='auxiliarycoordinate0'``

            default: optional
                Return the value of the *default* parameter if the
                data axes have not been set.

                {{default Exception}}

        :Returns:

            `tuple`
                The removed keys of the domain axis constructs spanned
                by the data.

        **Examples**

        >>> f.del_data_axes(key='dimensioncoordinate2')
        ('domainaxis1',)

        >>> f.has_data_axes(key='auxiliarycoordinate0')
        False
        >>> f.has_data_axes(key='auxiliarycoordinate0', default='no axes')
        'no axes'

        """
        data_axes = self.constructs.get_data_axes(key, default=None)
        if data_axes is not None:
            self.constructs._del_data_axes(key)
            return data_axes

        if default is None:
            return default

        return self._default(
            default,
            message=(
                f"{self.__class__.__name__!r} has no data axes for "
                f"construct {key!r}"
            ),
        )

    def has_data_axes(self, key):
        """Whether the axes spanned by the construct data have been set.

        Specifically, whether the domain axis constructs spanned by the
        data of a metadata construct have been set.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `del_data_axes`, `get_data_axes`, `set_data_axes`

        :Parameters:

            key: `str`
                Specify a metadata construct.

                *Parameter example:*
                  ``key='domainancillary1'``

        :Returns:

            `bool`
                True if domain axis constructs that span the data been
                set, otherwise False.

        **Examples**

        >>> f.set_data_axes(['domainaxis0', 'domainaxis1'])
        >>> f.get_data_axes()
        ('domainaxis0', 'domainaxis1')
        >>> f.del_data_axes()
        ('domainaxis0', 'domainaxis1')
        >>> print(f.del_dataxes(None))
        None
        >>> print(f.get_data_axes(default=None))
        None

        """
        return self.get_data_axes(key, default=None) is not None

    def set_data_axes(self, axes, key):
        """Sets domain axis constructs spanned by the construct data.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `data`, `del_data_axes`, `get_data`, `get_data_axes`,
                     `has_data_axes`

        :Parameters:

             axes: sequence of `str`
                The identifiers of the domain axis constructs spanned
                by the data of the field or of a metadata construct.

                *Parameter example:*
                  ``axes='domainaxis1'``

                *Parameter example:*
                  ``axes=['domainaxis1']``

                *Parameter example:*
                  ``axes=['domainaxis1', 'domainaxis0']``

             key: `str`
                Specify a metadata construct.

                *Parameter example:*
                  ``key='domainancillary1'``

        :Returns:

            `None`

        **Examples**

        >>> f.set_data_axes(['domainaxis0', 'domainaxis1'])
        >>> f.get_data_axes()
        ('domainaxis0', 'domainaxis1')
        >>> f.del_data_axes()
        ('domainaxis0', 'domainaxis1')
        >>> print(f.del_dataxes(None))
        None
        >>> print(f.get_data_axes(None))
        None

        """
        self.constructs._set_construct_data_axes(key=key, axes=axes)



from ..functions import abspath


class Files:
    """Mixin class for storing simple netCDF elements.

    .. versionadded:: (cfdm) 1.10.0.1

    """

    def _initialise_original_filenames(self, source=None):
        """Helps to initialise original file names.

        Call this from inside the `!__init__` method of a class that
        inherits from this mixin class.

        :Parameters:

            {{init source: optional}}

        :Returns:

            `None`

        **Examples**

        >>> f._initialise_original_filenames(source)

        """
        if source is not None:
            # Note: Getting and setting the component directly (as
            #       opposed to using the `_original_filenames` methods
            #       of 'source' and 'self') should improve the
            #       performance of `self.copy`.
            try:
                filenames = source._get_component("original_filenames", None)
            except AttributeError:
                pass
            else:
                if filenames:
                    self._set_component(
                        "original_filenames", filenames, copy=False
                    )

    def _original_filenames(self, define=None, update=None, clear=False):
        """The names of files containing the original data and metadata.

        {{original filenames}}

        .. versionadded:: (cfdm) 1.10.0.1

        .. seealso:: `get_original_filenames`

        :Parameters:

            {{define: (sequence of) `str`, optional}}

            {{update: (sequence of) `str`, optional}}

            {{clear: `bool` optional}}

        :Returns:

            `set` or `None`
                {{Returns original filenames}}

                If the *define* or *update* parameter is set then
                `None` is returned.

        **Examples**

        >>> d = {{package}}.{{class}}(9)
        >>> d._original_filenames()
        ()
        >>> d._original_filenames(define="file1.nc")
        >>> d._original_filenames()
        ('/data/user/file1.nc',)
        >>> d._original_filenames(update=["file1.nc"])
        >>> d._original_filenames()
        ('/data/user/file1.nc',)
        >>> d._original_filenames(update="file2.nc")
        >>> d._original_filenames()
        ('/data/user/file1.nc', '/data/user/file2.nc')
        >>> d._original_filenames(define="file3.nc")
        >>> d._original_filenames()
        ('/data/user/file3.nc',)
        >>> d._original_filenames(clear=True)
        >>> d._original_filenames()
        ()

        """
        filenames = None

        if define:
            # Replace the existing collection of original file names
            if isinstance(define, str):
                define = (define,)

            filenames = tuple([abspath(name) for name in define])

        if update:
            # Add new original file names to the existing collection
            if define is not None:
                raise ValueError(
                    "Can't set the 'define' and 'update' parameters "
                    "at the same time"
                )

            filenames = self._get_component("original_filenames", ())
            if isinstance(update, str):
                update = (update,)

            filenames += tuple([abspath(name) for name in update])

        if filenames:
            if len(filenames) > 1:
                filenames = tuple(set(filenames))

            self._set_component("original_filenames", filenames, copy=False)

        if define is not None or update is not None:
            if clear:
                raise ValueError(
                    "Can't set the 'clear' parameter with either of the "
                    "'define' and 'update' parameters"
                )

            # Return None, having potentially changed the file names
            return

        # Still here? Then return the existing original file names
        if clear:
            return set(self._del_component("original_filenames", ()))

        return set(self._get_component("original_filenames", ()))

    def get_original_filenames(self):
        """The names of files containing the original data and metadata.

        {{original filenames}}

        .. versionadded:: (cfdm) 1.10.0.1

        :Returns:

            `set`
                {{Returns original filenames}}

        """
        return self._original_filenames()



from . import core, mixin
from .decorators import _inplace_enabled, _inplace_enabled_define_and_cleanup


class CellConnectivity(
    mixin.NetCDFVariable,
    mixin.Topology,
    mixin.PropertiesData,
    mixin.Files,
    core.CellConnectivity,
):
    """A cell connectivity construct of the CF data model.

    A cell connectivity construct defines explicitly how cells
    arranged in two or three dimensions in real space but indexed by a
    single domain (discrete) axis are connected. Connectivity can only
    be provided when the domain axis construct also has a domain
    topology construct, and two cells can only be connected if they
    also have a topological relationship. For instance, the
    connectivity of two-dimensional face cells could be characterised
    by whether or not they have shared edges, where the edges are
    defined by connected nodes of the domain topology construct.

    The cell connectivity construct consists of an array recording the
    connectivity, and properties to describe the data. There must be a
    property indicating the condition by which the connectivity is
    derived from the domain topology. The array spans the domain axis
    construct with the addition of a ragged dimension. For each cell,
    the first element along the ragged dimension contains the unique
    identity of the cell, and the following elements contain in
    arbitrary order the identities of all the other cells to which the
    cell is connected. Note that the connectivity array for point
    cells is, by definition, equivalent to the array of the domain
    topology construct.

    When cell connectivity constructs are present they are considered
    to define the connectivity of the cells. Exactly the same
    connectivity information could be derived from the domain topology
    construct. Connectivity information inferred from inspection of
    any other constructs is not guaranteed to be the same.

    In CF-netCDF a cell topology construct can only be provided by a
    UGRID mesh topology variable. The construct array is supplied
    either indirectly by any of the UGRID variables that are used to
    define a domain topology construct, or directly by the UGRID
    "face_face_connectivity" variable (for face cells). In the direct
    case, the integer indices contained in the UGRID variable may be
    used as the cell identities, although the CF data model attaches
    no significance to the values other than the fact that some values
    are the same as others.

    Restricting the types of connectivity to those implied by the
    geospatial topology of the cells precludes connectivity derived
    from any other sources, but is consistent with UGRID encoding
    within CF-netCDF.

    See CF Appendix I "The CF Data Model".

    **NetCDF interface**

    {{netCDF variable}}

    .. versionadded:: (cfdm) 1.11.0.0

    """

    def __init__(
        self,
        connectivity=None,
        properties=None,
        data=None,
        source=None,
        copy=True,
        _use_data=True,
    ):
        """**Initialisation**

        :Parameters:

            {{init connectivity: `str`, optional}}

            {{init properties: `dict`, optional}}

                *Parameter example:*
                  ``properties={'long_name': 'face-face connectivity'}``

            {{init data: data_like, optional}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            connectivity=connectivity,
            properties=properties,
            data=data,
            source=source,
            copy=copy,
            _use_data=_use_data,
        )

        self._initialise_netcdf(source)
        self._initialise_original_filenames(source)

    def creation_commands(
        self,
        representative_data=False,
        namespace=None,
        indent=0,
        string=True,
        name="c",
        data_name="data",
        header=True,
    ):
        """Returns the commands to create the construct.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `{{package}}.Data.creation_commands`,
                     `{{package}}.Field.creation_commands`

        :Parameters:

            {{representative_data: `bool`, optional}}

            {{namespace: `str`, optional}}

            {{indent: `int`, optional}}

            {{string: `bool`, optional}}

            {{name: `str`, optional}}

            {{data_name: `str`, optional}}

            {{header: `bool`, optional}}

        :Returns:

            {{returns creation_commands}}

        """
        out = super().creation_commands(
            representative_data=representative_data,
            indent=0,
            namespace=namespace,
            string=False,
            name=name,
            data_name=data_name,
            header=header,
        )

        connectivity = self.get_connectivity(None)
        if connectivity is not None:
            out.append(f"{name}.set_connectivity({connectivity!r})")

        if string:
            indent = " " * indent
            out[0] = indent + out[0]
            out = ("\n" + indent).join(out)

        return out

    def dump(
        self,
        display=True,
        _omit_properties=None,
        _key=None,
        _level=0,
        _title=None,
        _axes=None,
        _axis_names=None,
    ):
        """A full description of the cell connectivity construct.

        Returns a description of all properties, including those of
        components, and provides selected values of all data arrays.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

        :Returns:

            {{returns dump}}

        """
        if _title is None:
            _title = "Cell Connectivity: " + self.identity(default="")

        return super().dump(
            display=display,
            _key=_key,
            _omit_properties=_omit_properties,
            _level=_level,
            _title=_title,
            _axes=_axes,
            _axis_names=_axis_names,
        )

    def identity(self, default=""):
        """Return the canonical identity.

        By default the identity is the first found from the following,
        in order:

        * The connectivity, preceded by ``'connectivity:'``.
        * The ``standard_name`` property.
        * The ``cf_role`` property, preceded by 'cf_role='.
        * The ``long_name`` property, preceded by 'long_name='.
        * The netCDF variable name, preceded by 'ncvar%'.
        * The value of the default parameter.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `identities`

        :Parameters:

            default: optional
                If no identity can be found then return the value of the
                default parameter.

        :Returns:

                The identity.

        """
        n = self.get_connectivity(None)
        if n is not None:
            return f"connectivity:{n}"

        n = self.get_property("standard_name", None)
        if n is not None:
            return n

        for prop in ("cf_role", "long_name"):
            n = self.get_property(prop, None)
            if n is not None:
                return f"{prop}={n}"

        n = self.nc_get_variable(None)
        if n is not None:
            return f"ncvar%{n}"

        return default

    def identities(self, generator=False, **kwargs):
        """Return all possible identities.

        The identities comprise:

        * The connectivity type, preceded by ``'connectivity:'``.
        * The ``standard_name`` property.
        * All properties, preceded by the property name and a colon,
          e.g. ``'long_name:Air temperature'``.
        * The netCDF variable name, preceded by ``'ncvar%'``.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `identity`

        :Parameters:

            generator: `bool`, optional
                If True then return a generator for the identities,
                rather than the default, a list.

            kwargs: optional
                Additional configuration parameters that may be used
                by subclasses.

        :Returns:

            `list` or `generator`
                The identities.

        """
        n = self.get_connectivity(None)
        if n is not None:
            pre = ((f"connectivity:{n}",),)
            pre0 = kwargs.pop("pre", None)
            if pre0:
                pre = tuple(pre0) + pre

            kwargs["pre"] = pre

        return super().identities(generator=generator, **kwargs)

    @_inplace_enabled(default=False)
    def normalise(
        self, start_index=0, remove_empty_columns=False, inplace=False
    ):
        """Normalise the data values.

        Normalised data is in a form that is suitable for creating a
        CF-netCDF UGRID "face_face_connectivity" variable.

        Normalisation does not change the logical content of the
        data. It converts the data so that the set of values comprises
        all of the integers in the range ``[0, N-1]`` (if the
        *start_index* parameter is ``0``), or ``[1, N]`` (if
        *start_index* is ``1``), where ``N`` is the number of cells.

        .. versionadded:: (cfdm) 1.11.0.0

        :Parameters:

            start_index: `int`, optional
                The start index for the data values in the normalised
                data. Must be ``0`` (the default) or ``1`` for zero-
                or one-based indices respectively.

            remove_empty_columns: `bool`, optional
                If True then remove any array columns that are
                entirely missing data. By default these are retained.

            {{inplace: `bool`, optional}}

        :Returns:

            `{{class}}` or `None`
                The normalised cell connectivity construct, or `None`
                if the operation was in-place.

        **Examples*

        >>> data = {{package}}.Data(
        ...   [[4, 1, 10, 125], [1, 4, -99, -99], [125, 4, -99, -99]]
        ... ).masked_values(-99)
        >>> c = {{package}}.{{class}}(cell='point', data=data)
        >>> print(c.array)
        [[4 1 10 125]
         [1 4 -- --]
         [125 4 -- --]]
        >>> print(c.normalise().array)
        [[0 1 2]
         [1 0 --]
         [2 0 --]]
        >>> print(c.normalise(start_index=1).array)
        [[1 2 3]
         [2 1 --]
         [3 1 --]]

        """
        if start_index not in (0, 1):
            raise ValueError("The 'start_index' parameter must be 0 or 1")

        d = _inplace_enabled_define_and_cleanup(self)
        data = self._normalise_cell_ids(
            d.array, start_index, remove_empty_columns
        )
        d.set_data(data, copy=False)
        return d



from .abstract import Array


class SparseArray(Array):
    """An underlying sparse array.

    The sparse array is assumed to have the same API as `scipy` sparse
    arrays.

    .. versionadded:: (cfdm) 1.11.0.0

    """

    def __init__(self, array=None, source=None, copy=True):
        """**Initialisation**

        :Parameters:

            array: `numpy.ndarray`
                The numpy array.

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(source=source, copy=copy)

        if source is not None:
            try:
                array = source._get_component("array", None)
            except AttributeError:
                array = None

        self._set_component("array", array, copy=False)
        if array is not None:
            self._set_component("dtype", array.dtype, copy=False)
            self._set_component("shape", array.shape, copy=False)

    @property
    def array(self):
        """Return an independent numpy array containing the data.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `sparse_array`

        :Returns:

            `numpy.ndarray`
                An independent numpy array of the data.

        **Examples**

        >>> isinstance(a.array, numpy.ndarray)
        True

        """
        return self._get_component("array").toarray()

    @property
    def sparse_array(self):
        """Return an independent `scipy` sparse array of the data.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `array`

        :Returns:

                An independent `scipy` sparse array of the data.

        **Examples**

        >>> from scipy.sparse import issparse
        >>> issparse(a.sparse_array)
        True

        """
        return self._get_component("array").copy()

    def to_memory(self):
        """Bring data on disk into memory.

        .. versionadded:: (cfdm) 1.11.0.0

        :Returns:

                Returns the contained sparse array, which is already
                in memory.

        """
        return self



import numpy as np

from .abstract import MeshSubarray


class BoundsFromNodesSubarray(MeshSubarray):
    """A subarray of cell bounds defined by UGRID node coordinates.

    A subarray describes a unique part of the uncompressed array.

    The UGRID node coordinates contain the locations of the nodes of
    the domain topology. In UGRID, the bounds of edge or face cells
    may be defined by the these locations in conjunction with a
    mapping from from each cell boundary vertex to its corresponding
    coordinate value.

    .. versionadded:: (cfdm) 1.11.0.0

    .. seealso:: `BoundsFromNodesArray`

    """

    def __init__(
        self,
        data=None,
        indices=None,
        shape=None,
        compressed_dimensions=None,
        node_coordinates=None,
        start_index=None,
        cell_dimension=None,
        source=None,
        copy=True,
        context_manager=None,
    ):
        """**Initialisation**

        :Parameters:

            data: array_like
                A 2-d integer array that contains indices that map
                each cell boundary vertex to its corresponding
                position in the 1-d *node_coordinates* array, as found
                in a UGRID "edge_node_connectivity" or
                "face_node_connectivity" variable. This array contains
                the mapping for all subarrays.

            indices: `tuple` of `slice`
                The indices of *data* that define this subarray.

            shape: `tuple` of `int`
                The shape of the uncompressed subarray.

            {{init compressed_dimensions: `dict`}}

                *Parameter example:*
                  ``{1: (1,)}``

            node_coordinates: array_like
                A 1-d array that contains a coordinate location for
                each mesh node, as found in a UGRID "node_coordinates"
                variable. This array contains the node coordinates for
                all subarrays.

            {{init start_index: `int`}}

            {{init cell_dimension: `int`}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

            context_manager: function, optional
                A context manager that provides a runtime context for
                the conversion of *data* and *node_coordinates* to
                `numpy` arrays.

        """
        super().__init__(
            data=data,
            indices=indices,
            shape=shape,
            compressed_dimensions=compressed_dimensions,
            start_index=start_index,
            cell_dimension=cell_dimension,
            source=source,
            copy=copy,
            context_manager=context_manager,
        )

        if source is not None:
            try:
                node_coordinates = source._get_component(
                    "node_coordinates", None
                )
            except AttributeError:
                node_coordinates = None

        if node_coordinates is not None:
            self._set_component(
                "node_coordinates", node_coordinates, copy=False
            )

    def __getitem__(self, indices):
        """Return a subspace of the uncompressed data.

        x.__getitem__(indices) <==> x[indices]

        Returns a subspace of the uncompressed data as an independent
        numpy array.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        node_connectivity = self._select_data(check_mask=True)

        if np.ma.isMA(node_connectivity):
            node_indices = node_connectivity.compressed()
            u = np.ma.masked_all(self.shape, dtype=self.dtype)
            u[~node_connectivity.mask] = self._select_node_coordinates(
                node_indices
            )
        else:
            node_indices = node_connectivity.flatten()
            u = self._select_node_coordinates(node_indices)
            u = u.reshape(self.shape)

        if indices is Ellipsis:
            return u

        return u[indices]

    def _select_node_coordinates(self, node_indices):
        """Select node coordinates.

        .. versionadded:: (cfdm) 1.11.0.0

        .. seealso:: `_select_data`

        :Parameters:

            node_indices: `numpy.array`
                Indices to the node coordinates array. Must be a 1-d
                integer array. `start_index` will be automatically
                subtracted from the indices prior to the selection of
                node coordinates.

        :Returns:

           `numpy.ndarray`
               The node coordinates that correspond to the
               *node_indices*.

        """
        start_index = self.start_index
        if start_index:
            node_indices = node_indices - start_index

        return self._asanyarray(
            self.node_coordinates, node_indices, check_mask=False
        )

    @property
    def dtype(self):
        """The data-type of the uncompressed data.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        return self.node_coordinates.dtype

    @property
    def node_coordinates(self):
        """The coordinates representing the node locations.

        .. versionadded:: (cfdm) 1.11.0.0

        """
        return self._get_component("node_coordinates")



import datetime
import faulthandler
import os
import unittest

import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


class DomainAncillaryTest(unittest.TestCase):
    """Unit test for the DomainAncillary class."""

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')

        self.filename = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "test_file.nc"
        )

    def test_DomainAncillary__repr__str__dump(self):
        """Test all means of DomainAncillary inspection."""
        f = cfdm.read(self.filename)[0]
        domain_ancillaries = f.domain_ancillaries("ncvar%a")

        x = domain_ancillaries.value()
        _ = repr(x)
        _ = str(x)
        self.assertIsInstance(x.dump(display=False), str)

        self.assertIsInstance(
            x.dump(display=False, _key=domain_ancillaries.key()), str
        )

        x.nc_del_variable()
        self.assertIsInstance(x.dump(display=False), str)

    def test_DomainAncillary_bounds(self):
        """Test the bounds keyword argument to DomainAncillary."""
        f = cfdm.read(self.filename)[0]

        a = f.auxiliary_coordinates("latitude").value()
        cfdm.DomainAncillary(bounds=a)

    def test_DomainAncillary_properties(self):
        """Test the property access methods of DomainAncillary."""
        f = cfdm.read(self.filename)[0]
        x = f.domain_ancillaries("ncvar%a").value()

        x.set_property("long_name", "qwerty")

        self.assertEqual(x.get_property("long_name"), "qwerty")
        self.assertEqual(x.del_property("long_name"), "qwerty")
        self.assertIsNone(x.get_property("long_name", None))
        self.assertIsNone(x.del_property("long_name", None))

    def test_DomainAncillary_insert_dimension(self):
        """Test the `insert_dimension` DomainAncillary method."""
        f = cfdm.read(self.filename)[0]
        d = f.dimension_coordinates("grid_longitude").value()
        x = cfdm.DomainAncillary(source=d)

        self.assertEqual(x.shape, (9,))
        self.assertEqual(x.bounds.shape, (9, 2))

        y = x.insert_dimension(0)
        self.assertEqual(y.shape, (1, 9))
        self.assertEqual(y.bounds.shape, (1, 9, 2), y.bounds.shape)

        x.insert_dimension(-1, inplace=True)
        self.assertEqual(x.shape, (9, 1))
        self.assertEqual(x.bounds.shape, (9, 1, 2), x.bounds.shape)

    def test_DomainAncillary_transpose(self):
        """Test the transpose DomainAncillary method."""
        f = cfdm.read(self.filename)[0]
        a = f.auxiliary_coordinates("longitude").value()
        bounds = cfdm.Bounds(
            data=cfdm.Data(np.arange(9 * 10 * 4).reshape(9, 10, 4))
        )
        a.set_bounds(bounds)
        x = cfdm.DomainAncillary(source=a)

        self.assertEqual(x.shape, (9, 10))
        self.assertEqual(x.bounds.shape, (9, 10, 4))

        y = x.transpose()
        self.assertEqual(y.shape, (10, 9))
        self.assertEqual(y.bounds.shape, (10, 9, 4), y.bounds.shape)

        x.transpose([1, 0], inplace=True)
        self.assertEqual(x.shape, (10, 9))
        self.assertEqual(x.bounds.shape, (10, 9, 4), x.bounds.shape)

    def test_DomainAncillary_squeeze(self):
        """Test the squeeze DomainAncillary method."""
        f = cfdm.read(self.filename)[0]
        a = f.auxiliary_coordinates("longitude").value()
        bounds = cfdm.Bounds(
            data=cfdm.Data(np.arange(9 * 10 * 4).reshape(9, 10, 4))
        )
        a.set_bounds(bounds)
        x = cfdm.DomainAncillary(source=a)

        x.insert_dimension(1, inplace=True)
        x.insert_dimension(0, inplace=True)

        self.assertEqual(x.shape, (1, 9, 1, 10))
        self.assertEqual(x.bounds.shape, (1, 9, 1, 10, 4))

        y = x.squeeze()
        self.assertEqual(y.shape, (9, 10))
        self.assertEqual(y.bounds.shape, (9, 10, 4), y.bounds.shape)

        x.squeeze(2, inplace=True)
        self.assertEqual(x.shape, (1, 9, 10))
        self.assertEqual(x.bounds.shape, (1, 9, 10, 4), x.bounds.shape)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)



class DeprecationError(Exception):
    """Deprecation error."""

    pass


class NumpyArray:
    """An underlying `numpy` array.

    Deprecated at version 1.12.0.0 and is no longer available. Use
    `numpy` instead.

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(self, *args, **kwargs):
        """**Initialisation**"""
        raise DeprecationError(
            f"{self.__class__.__name__} was deprecated at version 1.12.0.0 "
            "and is no longer available. Use numpy instead."
        )



"""Configuration for netCDF group flattening.

.. versionadded:: (cfdm) 1.11.2.0

Portions of this code were adapted from the `netcdf_flattener`
library, which carries the following Apache 2.0 License:

Copyright (c) 2020 EUMETSAT

Licensed to the Apache Software Foundation (ASF) under one or more
contributor license agreements. The ASF licenses this file to you
under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy
of the License at http://www.apache.org/licenses/LICENSE-2.0.

"""

from dataclasses import dataclass

# Maximum length of name after which it is replaced with its hash
max_name_len = 256

# Separator for groups in the input dataset
group_separator = "/"

# Replacment for 'group_separator' in flattened names
flattener_separator = "__"

# Name prefix when reference can't be resolved. Only used if
# 'lax_mode=True' in `flatten`.
ref_not_found_error = "REF_NOT_FOUND"

# NetCDF global attribute in the flattened dataset containing the
# mapping of flattened attribute names to grouped attribute names
flattener_attribute_map = "_flattener_attribute_map"

# NetCDF global attribute in the flattened dataset containing the
# mapping of flattened dimension names to grouped attribute names
flattener_dimension_map = "_flattener_dimension_map"

# NetCDF global attribute in the flattened dataset containing the
# mapping of flattened variable names to grouped attribute names
flattener_variable_map = "_flattener_variable_map"


@dataclass()
class FlatteningRules:
    """Define the flattening rules for a netCDF attribute.

    For a named netCDF attribute, the rules a define how the contents
    of the attribute are flattened. For instance, it has to be defined
    that the ``ancillary_variables`` attribute contains the names of
    other netCDF variables.

    .. versionadded:: (cfdm) 1.11.2.0

    """

    # name: The name of attribute containing the reference to be
    #       flattened
    name: str
    # ref_to_dim: Positive integer if contains references to
    #             dimensions. If ref_to_dim and ref_to_var are both
    #             positive then the rule with the greater value is
    #             tested first.
    ref_to_dim: int = 0
    # ref_to_var: Positive integer if contains references to
    #             variables. If ref_to_dim and ref_to_var are both
    #             positive then the rule with the greater value is
    #             tested first.
    ref_to_var: int = 0
    # resolve_key: True if 'keys' have to be resolved in 'key1: value1
    #              key2: value2 value3' or 'key1 key2'
    resolve_key: bool = False
    # resolve_value: True if 'values' have to be resolved in 'key1:
    #                value1 key2: value2 value3'
    resolve_value: bool = False
    # stop_at_local_apex: True if upward research in the hierarchy has
    #                     to stop at local apex.
    stop_at_local_apex: bool = False
    # accept_standard_names: True if any standard name is valid in
    #                        place of references (in which case no
    #                        exception is raised if a reference cannot
    #                        be resolved, and the standard name is
    #                        used in place)
    accept_standard_names: bool = False
    # limit_to_scalar_coordinates: True if references to variables are
    #                              only resolved if present as well in
    #                              the 'coordinates' attributes of the
    #                              variable, and they are scalar.
    limit_to_scalar_coordinates: bool = False


# --------------------------------------------------------------------
# Set the flattening rules for named CF attributes
# --------------------------------------------------------------------
flattening_rules = {
    attr.name: attr
    for attr in (
        # ------------------------------------------------------------
        # Coordinates
        # ------------------------------------------------------------
        FlatteningRules(
            name="coordinates",
            ref_to_var=1,
            resolve_key=True,
            stop_at_local_apex=True,
        ),
        FlatteningRules(name="bounds", ref_to_var=1, resolve_key=True),
        FlatteningRules(name="climatology", ref_to_var=1, resolve_key=True),
        # ------------------------------------------------------------
        # Cell methods
        # ------------------------------------------------------------
        FlatteningRules(
            name="cell_methods",
            ref_to_dim=2,
            ref_to_var=1,
            resolve_key=True,
            accept_standard_names=True,
            limit_to_scalar_coordinates=True,
        ),
        # ------------------------------------------------------------
        # Cell measures
        # ------------------------------------------------------------
        FlatteningRules(
            name="cell_measures", ref_to_var=1, resolve_value=True
        ),
        # ------------------------------------------------------------
        # Coordinate references
        # ------------------------------------------------------------
        FlatteningRules(
            name="formula_terms", ref_to_var=1, resolve_value=True
        ),
        FlatteningRules(
            name="grid_mapping",
            ref_to_var=1,
            resolve_key=True,
            resolve_value=True,
        ),
        # ------------------------------------------------------------
        # Ancillary variables
        # ------------------------------------------------------------
        FlatteningRules(
            name="ancillary_variables", ref_to_var=1, resolve_key=True
        ),
        # ------------------------------------------------------------
        # Compression by gathering
        # ------------------------------------------------------------
        FlatteningRules(name="compress", ref_to_dim=1, resolve_key=True),
        # ------------------------------------------------------------
        # Discrete sampling geometries
        # ------------------------------------------------------------
        FlatteningRules(
            name="instance_dimension", ref_to_dim=1, resolve_key=True
        ),
        FlatteningRules(
            name="sample_dimension", ref_to_dim=1, resolve_key=True
        ),
        # ------------------------------------------------------------
        # Domain variables
        # ------------------------------------------------------------
        FlatteningRules(name="dimensions", ref_to_dim=1, resolve_key=True),
        # ------------------------------------------------------------
        # Aggregation variables
        # ------------------------------------------------------------
        FlatteningRules(
            name="aggregated_dimensions", ref_to_dim=1, resolve_key=True
        ),
        FlatteningRules(
            name="aggregated_data", ref_to_var=1, resolve_value=True
        ),
        # ------------------------------------------------------------
        # Cell geometries
        # ------------------------------------------------------------
        FlatteningRules(name="geometry", ref_to_var=1, resolve_key=True),
        FlatteningRules(name="interior_ring", ref_to_var=1, resolve_key=True),
        FlatteningRules(
            name="node_coordinates", ref_to_var=1, resolve_key=True
        ),
        FlatteningRules(name="node_count", ref_to_var=1, resolve_key=True),
        FlatteningRules(name="nodes", ref_to_var=1, resolve_key=True),
        FlatteningRules(
            name="part_node_count", ref_to_var=1, resolve_key=True
        ),
        # ------------------------------------------------------------
        # UGRID variables
        # ------------------------------------------------------------
        FlatteningRules(name="mesh", ref_to_var=1, resolve_key=True),
        FlatteningRules(
            name="edge_coordinates", ref_to_var=1, resolve_key=True
        ),
        FlatteningRules(
            name="face_coordinates", ref_to_var=1, resolve_key=True
        ),
        FlatteningRules(
            name="edge_node_connectivity", ref_to_var=1, resolve_key=True
        ),
        FlatteningRules(
            name="face_node_connectivity", ref_to_var=1, resolve_key=True
        ),
        FlatteningRules(
            name="face_face_connectivity", ref_to_var=1, resolve_key=True
        ),
        FlatteningRules(
            name="edge_face_connectivity", ref_to_var=1, resolve_key=True
        ),
        FlatteningRules(
            name="face_edge_connectivity", ref_to_var=1, resolve_key=True
        ),
        FlatteningRules(name="edge_dimension", ref_to_dim=1, resolve_key=True),
        FlatteningRules(name="face_dimension", ref_to_dim=1, resolve_key=True),
        # ------------------------------------------------------------
        # Compression by coordinate subsampling
        # ------------------------------------------------------------
        FlatteningRules(
            name="coordinate_interpolation",
            ref_to_var=1,
            resolve_key=True,
            resolve_value=True,
        ),
        FlatteningRules(
            name="tie_point_mapping",
            ref_to_dim=2,
            ref_to_var=1,
            resolve_key=True,
            resolve_value=True,
        ),
        FlatteningRules(
            name="interpolation_parameters", ref_to_var=1, resolve_value=True
        ),
    )
}
